import time

from Foundation import (NSAppleScript, NSAppleScriptErrorMessage, NSAppleScriptErrorBriefMessage,
                        NSAppleScriptErrorNumber)

from macuitest.lib.applescript_lib.aeconverter import AEConverter

ae_converter = AEConverter()


class AppleScriptError(Exception):
    """Indicates an AppleScript compilation/execution error."""

    def __init__(self, error_info):
        self._error_info = dict(error_info)

    def __repr__(self):
        return f'{self.__class__.__name__}({self._error_info})'

    @property
    def message(self) -> str:
        msg = self._error_info.get(NSAppleScriptErrorMessage)
        if not msg:
            msg = self._error_info.get(NSAppleScriptErrorBriefMessage, 'Script Error')
        return msg

    @property
    def number(self):
        """"int | None -- the error number, if given")"""
        return self._error_info.get(NSAppleScriptErrorNumber)


class AppleScriptWrapper:
    """Wrapper for AppleScript with a set of easy to use methods."""
    allowed_modifier_keys = ('control', 'shift', 'option', 'command')
    key_codes: dict = {'return': 36, 'esc': 53, 'delete': 51, 'left': 123, 'right': 124, 'up': 126, 'down': 125,
                       'tab': 48}
    __pass_as_key_code = {'"': (39, True), "'": (39, False)}

    def typewrite(self, phrase: str) -> None:
        """Type `phrase` with 0.005s delay between key presses."""
        for char in phrase:
            time.sleep(.005)
            if char in self.__pass_as_key_code:
                key_code, is_shift_required = self.__pass_as_key_code[char]
                self.send_keycode(key_code, 'shift') if is_shift_required else self.send_keycode(key_code)
            else:
                self.send_keystroke(char)

    def send_keystroke(self, phrase: str, *args) -> None:
        """Send keystroke event with the specified `phrase` (type it using AppleScript)."""
        self.__send_event('keystroke', phrase, *args)

    def send_keycode(self, key_code: int, *args) -> None:
        """Send keystroke event with the specified `key_code` (type it using AppleScript)."""
        self.__send_event('key code', key_code, *args)

    def __send_event(self, event_type: str, message: [str, int], *args):
        for modifier in args:
            assert modifier in self.allowed_modifier_keys
        _cmd = f'{event_type} "{message}"'
        if args:
            _cmd = f'{event_type} "{message}" using {{{", ".join([f"{modifier_key} down" for modifier_key in args])}}}'
        return self.tell_sys_events(_cmd)

    def tell_app_process(self, command: str, app_process: str):
        return self.execute(f'tell app "System Events" to tell application process "{app_process}" to {command}')

    def tell_app(self, app: str, command: str, ignoring_responses: bool = False):
        _tell_what = f'tell application "{app}" to {command}' if not ignoring_responses else f'ignoring application responses\ntell application "{app}" to {command}\nend ignoring'
        return self.execute(_tell_what)

    def tell_sys_events(self, command: str):
        return self.execute(f'tell application "System Events" to {command}')

    @staticmethod
    def execute(cmd: str):
        """Execute AppleScript command abd returns exitcode, stdout and stderr.
            :param str cmd: apple script
            :return: exitcode, stdout and stderr"""
        result, error = NSAppleScript.alloc().initWithSource_(cmd).executeAndReturnError_(None)
        if error:
            raise AppleScriptError(error)
        return ae_converter.unpack(result)


as_wrapper = AppleScriptWrapper()



import os
import time
from dataclasses import dataclass
from datetime import datetime
from types import MappingProxyType
from typing import Any, Union, Optional, Tuple

from macuitest.config.constants import CheckboxState, Frame, Point, DisclosureTriangleState, Region
from macuitest.lib import core
from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper, AppleScriptError
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.controllers.mouse import mouse
from macuitest.lib.operating_system.color_meter import ColorMeter
from macuitest.lib.operating_system.env import env


class BaseUIElement:
    """Default AppleScript UI element."""
    __slots__ = ('locator', 'process', '_frame')

    def __init__(self, locator, process):
        self.locator = locator
        self.process = process
        self._frame = None

    def __repr__(self):
        return f"<{self.__class__.__name__}|locator='{self.locator}', process='{self.process}'>"

    def wait_on_position(self, position: Point) -> bool:
        self.reset_frame()
        return self.frame.center == position

    def snapshot(self, margin: int = 4) -> str:
        self.reset_frame()
        _region = f'{self.frame.x1 - margin},{self.frame.y1 - margin},' \
                  f'{self.frame.width + margin * 2},{self.frame.height + margin * 2}'
        _where = f'{env.desktop}/scr_{datetime.now().strftime("%Y%m%d%H%M%S%f")}.png'
        os.system(f'screencapture -x -R"{_region}" "{_where}"')
        return _where

    def region(self, margin: int = 4):
        self.reset_frame()
        return Region(self.frame.x1 - margin, self.frame.y1 - margin, self.frame.x2 + margin, self.frame.y2 + margin)

    def most_common_color(self, ignore_colors: Optional[Tuple[str, ...]] = None):
        self.reset_frame()
        f = self.frame
        return ColorMeter().get_most_common_color(f.x1, f.x2, f.y1, f.y2, ignore_colors)

    def color(self, x_off: int = 0, y_off: int = 0) -> str:
        self.reset_frame()
        pixel_color = ColorMeter().get_color(Point(self.frame.center.x + x_off, self.frame.center.y + y_off))
        return pixel_color.replace('gray', 'grey')

    def scroll(self, clicks, x_off=0, y_off=0):
        self.__assert_visible()
        mouse.scroll(self.frame.center.x + x_off, self.frame.center.y + y_off, clicks, x_off, y_off)

    def doubleclick_mouse(self, x_off: int = 0, y_off: int = 0, duration: Union[int, float] = 0.15):
        self.__assert_visible()
        mouse.double_click(self.frame.center.x + x_off, self.frame.center.y + y_off, duration=duration)
        self.reset_frame()

    def rightclick_mouse(self, x_off=0, y_off=0, duration=.2) -> None:
        self.__assert_visible()
        mouse.right_click(self.frame.center.x + x_off, self.frame.center.y + y_off, duration)
        self.reset_frame()

    def click_mouse(self, x_off=0, y_off=0, hold_time=.3, duration=.2, pause=.3) -> None:
        if not self.wait_displayed(timeout=7):
            raise LookupError(self)
        mouse.click(self.frame.center.x + x_off, self.frame.center.y + y_off, hold_time, duration, pause)
        self.reset_frame()

    def hover_mouse(self, x_off: int = 0, y_off: int = 0, duration: float = .2) -> None:
        mouse.hover(self.frame.center.x + x_off, self.frame.center.y + y_off, duration=duration)

    def click(self, pause: float = 0.375) -> bool:
        """Perform click action the element."""
        self.__assert_visible()
        time.sleep(pause)
        self.__execute('click')
        time.sleep(0.2)
        return True

    def _select(self):
        """Click an element."""
        self.__assert_visible()
        return self.__execute('select')

    def _set_focus(self, value):
        """Make element focused."""
        self.__assert_visible()
        return self.__execute('set focused of', params=f'to "{value}"')

    def _show_context_menu(self):
        self.__assert_visible()
        return self.__execute('perform action "AXShowMenu" of')

    def _set_value(self, value):
        """Set element value."""
        self.__assert_visible()
        return self.__execute('set value of', params=f'to "{value}"')

    def _get_value(self):
        """Get element value."""
        self.__assert_visible()
        return self.__execute('get value of')

    def _count_elements(self) -> int:
        """Count UI elements """
        self.__assert_visible()
        return self.__execute('count UI elements of')

    @property
    def _children(self):
        self.__assert_visible()
        return self._read_attribute('AXChildren')

    @property
    def _rows(self) -> int:
        self.__assert_visible()
        return self.__execute('count rows of')

    @property
    def frame(self) -> Frame:
        if self._frame is not None:
            return self._frame
        self.__assert_visible()
        _frame = [*self._read_attribute('AXPosition'), *self._read_attribute('AXSize')]
        x1, y1, width, height = _frame
        x2, y2 = x1 + width, y1 + height
        center = Point(int((x1 + width / 2)), int((y1 + height / 2)))
        self._frame = Frame(x1, y1, x2, y2, center, width, height)
        return self._frame

    def reset_frame(self):
        self._frame = None

    @property
    def title(self) -> str:
        self.__assert_visible()
        return self._read_attribute('AXTitle').strip()

    @property
    def description(self) -> str:
        self.__assert_visible()
        return self._read_attribute('AXDescription').strip()

    @property
    def value(self) -> str:
        self.__assert_visible()
        return self._read_attribute('AXValue')

    @property
    def help(self) -> str:
        self.__assert_visible()
        return self._read_attribute('AXHelp')

    @property
    def _placeholder(self) -> str:
        self.__assert_visible()
        return self._read_attribute('AXPlaceholderValue')

    def _is_enabled(self) -> bool:
        """Check whether element enabled"""
        self.__assert_visible()
        return self._read_attribute('AXEnabled')

    @property
    def did_vanish(self) -> bool:
        return self.wait_vanish()

    def wait_vanish(self, timeout: [int, float] = 5) -> bool:
        self.wait_displayed(timeout=0.3)
        return wait_condition(lambda: self.is_exists() is False, timeout=timeout)

    @property
    def is_visible(self) -> bool:
        return self.wait_displayed()

    def wait_displayed(self, timeout: Union[int, float] = 5):
        return wait_condition(self.is_exists, timeout=timeout)

    def _set_attribute(self, attribute, value) -> Any:
        return self.__execute(f'set value of attribute "{attribute}" of', params=f'to {value}')

    def _read_attribute(self, attribute) -> Any:
        try:
            return self.__execute(f'get value of attribute "{attribute}" of')
        except AppleScriptError:
            return None

    def is_exists(self) -> bool:
        try:
            return self.__execute(f'return exists')
        except AppleScriptError as e:
            if e.number == -10000:
                pass
            else:
                raise

    def __assert_visible(self):
        if not self.is_visible:
            raise LookupError(self)

    def __execute(self, command, params=''):
        """Execute a command.
               :param str command: The name of the command to _execute as a string.
               :param str params: Command parameters.
               :return str: Execution output."""
        _command = f'{command} {self.locator} {params}' if params else f'{command} {self.locator}'
        return as_wrapper.tell_app_process(command=_command, app_process=self.process)


class Button(BaseUIElement):
    @property
    def is_enabled(self):
        return wait_condition(self._is_enabled)

    def wait_enabled(self, timeout: int) -> bool:
        return wait_condition(self._is_enabled, timeout=timeout)

    @property
    def is_disabled(self):
        return wait_condition(lambda: self._is_enabled() is False)

    @property
    def value(self):
        return self._get_value()


class BusyIndicator(BaseUIElement):
    pass


class TextElement(BaseUIElement):
    def __repr__(self):
        try:
            return f"'{self.text}'"
        except (AppleScriptError, LookupError):
            return f"{self.__class__.__name__}(locator='{self.locator}', process='{self.process}')"

    @property
    def to_bytes(self) -> int:
        return core.convert_file_size_to_bytes(self.text)

    @property
    def characters_number(self):
        return self._read_attribute('AXNumberOfCharacters')

    @property
    def visible_characters_number(self):
        return list(self._read_attribute('AXVisibleCharacterRange'))[1]

    def set_text(self, value: str):
        self.wait_displayed()
        self._set_focus(True)
        self._set_value(value)

    def get_text(self) -> str:
        self.wait_displayed(timeout=1)
        return str(self._get_value())

    @property
    def is_fully_visible(self):
        self.wait_displayed(timeout=1)
        return self.characters_number == self.visible_characters_number

    def __eq__(self, other):
        return wait_condition(lambda: self.text == other, timeout=1)

    text = property(get_text, set_text)


class StaticText(TextElement):
    pass


class TextArea(TextElement):
    pass


class TextField(TextElement):
    @property
    def placeholder(self):
        return self._placeholder

    def value(self) -> str:
        return self._get_value()

    def set(self, value):
        self._set_value(value)

    def focus(self, value='true'):
        self._set_focus(value)

    @property
    def is_focused(self):
        return self._read_attribute('AXFocused')

    @is_focused.setter
    def is_focused(self, value: bool):
        self._set_focus(value)

    text = property(value, set)


class ComboBox(BaseUIElement):
    def value(self) -> str:
        return self._get_value()

    def set_value(self, value):
        self.wait_displayed()
        time.sleep(0.5)
        self._set_value(value)

    text = property(value, set_value)


class Checkbox(BaseUIElement):
    @property
    def is_enabled(self):
        return wait_condition(self._is_enabled, timeout=3)

    @property
    def is_disabled(self):
        return wait_condition(lambda: self._is_enabled() is False)

    def wait_state(self, state):
        return wait_condition(lambda: self.get_state() == state)

    def set_state(self, state):
        target_state = CheckboxState.states.get(state)
        for _ in range(2):
            if self.get_state() != target_state:
                self.click()
        assert self.state == target_state, f'Cannot set checkbox to state: "{state}"'

    def get_state(self):
        return self._get_value()

    state = property(get_state, set_state)


class DisclosureTriangle(BaseUIElement):
    def get_state(self):
        return self._get_value()

    def set_state(self, state: DisclosureTriangleState):
        attempts_left = 2
        while self.state != state and attempts_left:
            self.click()
            attempts_left -= 1
        assert self.state == state, f'Cannot set disclosure triangle to state: "{state}"'

    state = property(get_state, set_state)


class MenuItem(BaseUIElement):
    @property
    def is_enabled(self):
        return wait_condition(self._is_enabled)


class MenuBarItem(BaseUIElement):
    pass


class Menu(BaseUIElement):
    pass


class Group(BaseUIElement):
    @property
    def elements_number(self):
        return self._count_elements()


class Row(BaseUIElement):
    def select(self, pause: float = 0.3):
        self._select()
        time.sleep(pause)

    @property
    def is_selected(self) -> bool:
        return self._read_attribute('AXSelected')


class ScrollArea(BaseUIElement):
    pass


class Table(BaseUIElement):
    @property
    def rows_number(self) -> int:
        return self._rows


class Image(BaseUIElement):
    @property
    def label(self):
        return self.description


class List(BaseUIElement):
    @property
    def elements_number(self):
        return self._count_elements()


class Outline(BaseUIElement):
    @property
    def elements_number(self):
        return self._count_elements()


class Popover(BaseUIElement):
    pass


class Sheet(BaseUIElement):
    pass


class Window(BaseUIElement):
    def __init__(self, locator, process):
        super().__init__(locator, process)

    def set_minimized(self, value):
        converted = {True: 'true', False: 'false'}.get(value)
        self._set_attribute('AXMinimized', converted)

    def is_minimized(self):
        return self._read_attribute('AXMinimized')

    def set_full_screen(self, value):
        converted = {True: 'true', False: 'false'}.get(value)
        self._set_attribute('AXFrontmost', converted)

    def is_full_screen(self):
        return self._read_attribute('AXFullScreen')

    @property
    def is_keyboard_focused(self) -> bool:
        return self._read_attribute('AXFocused')

    minimized = property(is_minimized, set_minimized)
    full_screen = property(is_full_screen, set_full_screen)


class WebView(BaseUIElement):
    @property
    def url(self):
        wait_condition(lambda: self._read_attribute('AXURL') is not None, timeout=60)
        wait_condition(lambda: self._read_attribute('AXURL').startswith('http'), timeout=30)
        return wait_condition(lambda: self._read_attribute('AXURL'), timeout=30)


@dataclass(frozen=True)
class Elements:
    all: MappingProxyType = MappingProxyType(
        {'UI element': BaseUIElement, 'button': Button, 'pop up button': Button, 'radio button': Button,
            'busy indicator': BusyIndicator, 'static text': StaticText, 'text area': TextArea, 'text field': TextField,
            'checkbox': Checkbox, 'combo box': ComboBox, 'menu item': MenuItem, 'menu button': MenuItem,
            'menu bar item': MenuBarItem, 'menu': Menu, 'group': Group, 'row': Row, 'scroll area': ScrollArea,
            'table': Table, 'image': Image, 'list': List, 'outline': Outline, 'pop over': Popover, 'popover': Popover,
            'sheet': Sheet, 'window': Window, })


class ASElement:
    """AppleScript element factory."""
    __slots__ = ()
    window: str = 'of window 1'

    def __new__(cls, locator: str, process: str) \
            -> Union[BaseUIElement, Button, BusyIndicator, TextArea, TextField, Checkbox, ComboBox, MenuItem,
                     Group, Row, ScrollArea, Table, Image, List, Outline, Popover, Sheet, Window]:
        locator = locator.replace('of of', 'of')
        return cls.__define_locator_type(locator)(locator, process)

    @staticmethod
    def __define_locator_type(locator: str):
        _l = locator.split(' of')[0]
        _l = (''.join((i for i in _l if not i.isdigit())) if _l[-1].isdigit() else _l.split('"')[0]).strip()
        return Elements.all.get(_l, BaseUIElement)



"""This module implements a basic color meter. It finds the value of a colour on given coordinates."""
import os
from collections import Counter
from multiprocessing import Pool
from typing import Optional, Tuple

import webcolors
from PIL import Image

from macuitest.config.constants import Point
from macuitest.lib.ui.monitor import monitor


class ColorMeter:
    def __init__(self):
        self.__pixels = None

    def get_most_common_color(self, x1: int, x2: int, y1: int, y2: int,
                              ignore_colors: Optional[Tuple[str, ...]] = None):
        pool = Pool(os.cpu_count())
        pixels = [self._pixels[x, y] for x in range(x1, x2) for y in range(y1, y2)]
        colors = pool.map(self.get_closest_color, pixels)
        if ignore_colors:
            colors = (c for c in colors if c not in ignore_colors)
        return Counter(colors).most_common()[0][0]

    def get_color(self, point: Point) -> str:
        return self.get_closest_color(self._pixels[point.x, point.y]).replace('gray', 'grey')

    @staticmethod
    def get_closest_color(pixel) -> str:
        min_colours = {}
        for key, name in webcolors.CSS3_HEX_TO_NAMES.items():
            r_c, g_c, b_c = webcolors.hex_to_rgb(key)
            min_colours[((r_c - pixel[0]) ** 2 + (g_c - pixel[1]) ** 2 + (b_c - pixel[2]) ** 2)] = name
        return min_colours[min(min_colours.keys())]

    @property
    def _pixels(self):
        if self.__pixels is None:
            size = (monitor.size.width, monitor.size.height)
            self.__pixels = Image.frombytes('RGB', size, monitor.bytes, 'raw', 'BGRX').load()
            return self.__pixels
        return self.__pixels



from typing import Union, Optional

import cv2
import numpy

from macuitest.config.constants import Point, Region
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.controllers.mouse import mouse
from macuitest.lib.ui.monitor import monitor
from macuitest.lib.ui.opencv_wrapper import opencv_wrapper


class UIElementNotFoundOnScreen(Exception):
    """Thrown when a pattern is not found on the screen."""
    pass


class UIElement:
    """Represent a visible user interface element. Based on automated pattern recognition algorithm (OpenCV)."""

    def __init__(self, scr_path: str, sim: float = 0.85):
        self.pattern = scr_path.strip()
        self.similarity = sim
        self.image, self.width, self.height = None, None, None
        self._center = None
        self.load_image()

    def __repr__(self):
        return f"{self.__class__.__name__}(pattern='..{self.short_name}', sim={self.similarity})"

    @property
    def short_name(self):
        return self.pattern.split('media')[-1]

    def paste(self, x_off=0, y_off=0, phrase=''):
        mouse.paste(self.center.x, self.center.y, x_off, y_off, phrase=phrase)
        self._center = None

    def double_click(self, x_off=0, y_off=0):
        mouse.double_click(self.center.x, self.center.y, x_off, y_off)
        self._center = None

    def right_click(self, x_off=0, y_off=0):
        mouse.right_click(self.center.x, self.center.y, x_off, y_off)
        self._center = None

    def click(self, x_off=0, y_off=0, hold=.3):
        mouse.click(self.center.x, self.center.y, x_off, y_off, hold=hold)
        self._center = None

    def hover_above(self, x_off=0, y_off=0, duration=.2):
        mouse.hover(self.center.x, self.center.y, x_off, y_off, duration=duration)
        self._center = None

    @property
    def is_visible(self):
        return self.wait_displayed() is not None

    @property
    def center(self):
        match = self._center or self.wait_displayed()
        if not match:
            raise UIElementNotFoundOnScreen(self.short_name)
        return Point(match.x + self.width / 2, match.y + self.height / 2)

    def wait_displayed(self, timeout=5, region: Optional[Region] = None) -> Union[None, Point]:
        match = wait_condition(
            lambda: opencv_wrapper.find_pattern(self.image, self.similarity, region), timeout=timeout)
        if isinstance(match, Point):
            self._center = match
            return match

    def wait_vanish(self, timeout=30, region: Optional[Region] = None) -> bool:
        return wait_condition(
            lambda: opencv_wrapper.find_pattern(self.image, self.similarity, region) is None,  timeout=timeout)

    def load_image(self) -> numpy.ndarray:
        image = cv2.imread(self.pattern, 0)
        if image is None:
            return numpy.empty(2)
        height, width = image.shape
        if monitor.is_retina:
            height, width = height / 2, width / 2
        self.image, self.width, self.height = image, width, height



from ApplicationServices import (kAXErrorActionUnsupported, kAXErrorAPIDisabled, kAXErrorAttributeUnsupported,
                                 kAXErrorCannotComplete, kAXErrorFailure, kAXErrorIllegalArgument,
                                 kAXErrorInvalidUIElement, kAXErrorInvalidUIElementObserver,
                                 kAXErrorNotificationAlreadyRegistered, kAXErrorNotificationNotRegistered,
                                 kAXErrorNotificationUnsupported, kAXErrorNotImplemented, kAXErrorNoValue,
                                 kAXErrorSuccess, )


class AXError(Exception):
    pass


class AXErrorUnsupported(AXError):
    pass


class AXErrorAPIDisabled(AXError):
    pass


class AXErrorInvalidUIElement(AXError):
    pass


class AXErrorCannotComplete(AXError):
    pass


class AXErrorNotImplemented(AXError):
    pass


class AXErrorIllegalArgument(AXError):
    pass


class AXErrorActionUnsupported(AXError):
    pass


class AXErrorNoValue(AXError):
    pass


class AXErrorFailure(AXError):
    pass


class AXErrorInvalidUIElementObserver(AXError):
    pass


class AXErrorNotificationUnsupported(AXError):
    pass


class AXErrorNotificationAlreadyRegistered(AXError):
    pass


class AXErrorNotificationNotRegistered(AXError):
    pass


class AXErrorAttributeUnsupported(AXError):
    pass


class AXErrorFactory:
    def __new__(cls, error_code):
        return {kAXErrorAPIDisabled: AXErrorAPIDisabled, kAXErrorInvalidUIElement: AXErrorInvalidUIElement,
                kAXErrorCannotComplete: AXErrorCannotComplete, kAXErrorNotImplemented: AXErrorNotImplemented,
                kAXErrorIllegalArgument: AXErrorIllegalArgument, kAXErrorNoValue: AXErrorNoValue,
                kAXErrorFailure: AXErrorFailure, kAXErrorInvalidUIElementObserver: AXErrorInvalidUIElementObserver,
                kAXErrorNotificationUnsupported: AXErrorNotificationUnsupported,
                kAXErrorNotificationAlreadyRegistered: AXErrorNotificationAlreadyRegistered,
                kAXErrorNotificationNotRegistered: AXErrorNotificationNotRegistered,
                kAXErrorAttributeUnsupported: AXErrorAttributeUnsupported,
                kAXErrorActionUnsupported: AXErrorActionUnsupported, }.get(error_code, AXErrorUnsupported)


def check_ax_error(error_code: int, error_messages) -> None:
    """ Return if code is kAXErrorSuccess.
         Raise an error with given message based on given error code.
         Defaults to AXErrorUnsupported for unknown codes.
            Args:
                error_code: the error code
                error_messages: mapping from error code to error message
    """
    if error_code == kAXErrorSuccess:
        return
    try:
        error_message = error_messages[error_code]
    except KeyError:
        if error_code == kAXErrorFailure:
            error_message = 'There is some sort of system memory failure'
        else:
            error_message = f'Unknown AX Error code: {error_code}'
    raise AXErrorFactory(error_code)(error_message)



import os
from typing import Any, Dict

import biplist

from macuitest.lib.core import wait_condition


class PropertyListMissing(Exception):
    pass


class PlistHelper:
    def __init__(self, property_list):
        self.plist = property_list

    def read_property(self, key) -> Any:
        return self.read_plist().get(key)

    def delete_property(self, key):
        _content = self.read_plist()
        try:
            _content.pop(key)
        except KeyError:
            return
        self.write_plist(_content)

    def set_property(self, key, value):
        _content = self.read_plist()
        _content[key] = value
        self.write_plist(_content)

    def write_plist(self, content):
        return biplist.writePlist(content, self.plist, binary=False)

    def read_plist(self) -> Dict[str, Any]:
        payload = dict()
        if not wait_condition(lambda: os.path.exists(self.plist), timeout=15):
            raise PropertyListMissing
        try:
            payload = biplist.readPlist(self.plist)
        except (biplist.NotBinaryPlistException, biplist.InvalidPlistException):
            raise
        finally:
            return payload



import glob
import os
from typing import List, Tuple

from macuitest.lib.operating_system.env import env


class CrashReporter:

    def __init__(self, executor):
        self.executor = executor

    def publish_reports(self, logger, crash_reports: List[str]):
        for report in crash_reports:
            self.__publish_crash_report(logger, report)

    def __publish_crash_report(self, logger, report: str):
        name = os.path.basename(report)
        self.executor.sudo(f'chmod 755 "{report}"')
        with open(report) as rep:
            logger.error(name, extra=dict(
                attachment={
                    "name": name,
                    "data": rep.read(),
                    "mime": "application/octet-stream",
                }))

    def archive_crash_reports(self, search_patterns: Tuple[str], where: str):
        for report in self.collect_reports(search_patterns):
            self.executor.execute(f'cp "{report}" "{where}"')
            self.executor.sudo(f'rm "{report}"')

    @staticmethod
    def collect_reports(search_patterns: Tuple[str, ...]) -> List[str]:
        """Collect crash reports grouped by a glob search pattern."""
        crash_reports = []
        for directory in (env.user_diag_reports, env.system_diag_reports):
            for search_pattern in search_patterns:
                for report in glob.glob(f'{directory}/{search_pattern}'):
                    crash_reports.append(report)
        return crash_reports



"""OpenCV (Open Source Computer Vision Library) wrapper.
OpenCV (Open Source Computer Vision Library) is an open source computer vision and machine learning software library.
OpenCV was built to provide a common infrastructure for computer vision applications and to accelerate the use
of machine perception in the commercial products. Being a BSD-licensed product, OpenCV makes it easy for businesses
to utilize and modify the code."""
from typing import Optional

import cv2

from macuitest.config.constants import Point, Region
from macuitest.lib.ui.monitor import monitor


class _WrapperOpenCV:
    algorithm = cv2.TM_CCOEFF_NORMED

    def __init__(self, display: monitor):
        self.display = display

    def find_pattern(self, pattern, sim: float, region: Optional[Region] = None) -> Optional[Point]:
        """Locate pattern on the screen and return its center."""
        region = Region(0, 0, self.display.size.width, self.display.size.height) if region is None else region
        desktop = cv2.cvtColor(self.display.screenshot, cv2.COLOR_BGR2GRAY)[region.y1:region.y2, region.x1:region.x2]
        _, similarity, _, position = cv2.minMaxLoc(cv2.matchTemplate(desktop, pattern, self.algorithm))
        if similarity > sim:
            denominator = 2 if self.display.is_retina else 1
            return Point((position[0] + region.x1) // denominator, (position[1] + region.y1) // denominator)


opencv_wrapper = _WrapperOpenCV(monitor)



import time
from typing import ClassVar

from macuitest.lib.operating_system.crash_reporter import CrashReporter
from macuitest.lib.operating_system.defaults import Defaults
from macuitest.lib.operating_system.disk_util import DiskUtil
from macuitest.lib.operating_system.file_manager import FileManager
from macuitest.lib.operating_system.hosts import Hosts
from macuitest.lib.operating_system.login_window import LoginWindow
from macuitest.lib.operating_system.memory_manager import MemoryManager
from macuitest.lib.operating_system.networking import Networking
from macuitest.lib.operating_system.security_manager import SecurityManager
from macuitest.lib.operating_system.service_manager import ServiceManager
from macuitest.lib.operating_system.shell_executor import ShellExecutor
from macuitest.lib.operating_system.system_information import SystemInformation
from macuitest.lib.operating_system.time_manager import TimeManager

__all__ = ['macos']


class MacOS:
    user_notification_center: ClassVar[str] = 'UserNotificationCenter'
    problem_reporter: ClassVar[str] = 'Problem Reporter'

    def __init__(self, shell_executor: ShellExecutor):
        self.shell_executor = shell_executor
        self.crash_reporter = CrashReporter(self.shell_executor)
        self.disk_util = DiskUtil(self.shell_executor)
        self.file_manager = FileManager(self.shell_executor)
        self.login_window = LoginWindow(self.shell_executor)
        self.memory_manager = MemoryManager(self.shell_executor)
        self.networking = Networking(self.shell_executor)
        self.security_manager = SecurityManager(self.shell_executor)
        self.service_manager = ServiceManager(self.shell_executor)
        self.sys_info = SystemInformation(self.shell_executor)
        self.time_manager = TimeManager(self.shell_executor)
        self.defaults = Defaults(self.shell_executor, self.service_manager)
        self.hosts = Hosts(self.shell_executor, self.file_manager)


macos = MacOS(shell_executor=ShellExecutor())



import os
import time
from typing import Dict, Union

from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper, AppleScriptError
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.applescript_element import ASElement
from macuitest.lib.operating_system.env import env
from macuitest.lib.operating_system.macos import macos
from macuitest.lib.operating_system.plist_helper import PlistHelper


class Application:
    def __init__(self, app_name: str, location: str = env.applications):
        self.name: str = app_name
        self.application_property_list: str = f'{location}/{self.name}.app/Contents/Info.plist'
        self.contents_reader: PlistHelper = PlistHelper(self.application_property_list)
        self.window = ASElement('window 1', process=self.name)

    def close_windows(self):
        try:
            return as_wrapper.tell_app(self.name, 'close every window')
        except AppleScriptError:
            pass

    def relaunch(self) -> None:
        self.quit()
        time.sleep(3)
        self.launch()

    def quit(self, pause: Union[int, float] = 2) -> None:
        time.sleep(pause)
        as_wrapper.tell_app(self.name, 'quit', ignoring_responses=True)
        if not macos.service_manager.wait_process_disappeared(self.name):
            macos.service_manager.kill_process(self.name)
        assert self.did_quit

    @property
    def did_quit(self) -> bool:
        return macos.service_manager.wait_process_disappeared(self.name)

    def launch(self) -> bool:
        macos.service_manager.launch_application_by_name(self.name)
        return self.window.wait_displayed(timeout=15)

    def activate(self):
        as_wrapper.tell_app(app=self.name, command='activate')
        assert self.frontmost is True

    def set_window_position(self, position: str = '{1210, 670}'):
        if self.frontmost:
            try:
                return as_wrapper.tell_app_process(f'set position of the front window to {position}', self.name)
            except AppleScriptError:
                pass

    def set_window_size(self, size: str = '{700, 400}'):
        if self.frontmost:
            try:
                return as_wrapper.tell_app_process(f'set size of the front window to {size}', self.name)
            except AppleScriptError:
                pass

    @property
    def version(self) -> int:
        return self.contents_reader.read_property('CFBundleShortVersionString')

    @property
    def build(self) -> int:
        return self.contents_reader.read_property('CFBundleVersion')

    @property
    def is_running(self) -> bool:
        return macos.service_manager.wait_process_appeared(self.name, timeout=2)

    @property
    def did_launch(self) -> bool:
        return macos.service_manager.wait_process_appeared(self.name, timeout=20)

    def set_frontmost(self, value: bool = True):
        converted = {True: 'true', False: 'false'}.get(value)
        self._set_attribute('AXFrontmost', converted)

    def is_frontmost(self) -> bool:
        return wait_condition(lambda: self._read_attribute('AXFrontmost'), timeout=3)

    def set_hidden(self, value: bool = True):
        converted = {True: 'true', False: 'false'}.get(value)
        self._set_attribute('AXHidden', converted)

    def is_hidden(self) -> bool:
        return wait_condition(lambda: self._read_attribute('AXHidden'), timeout=3)

    def _set_attribute(self, attribute, value):
        return self.__execute(f'set value of attribute "{attribute}"', params=f'to {value}')

    def _read_attribute(self, attribute):
        try:
            return self.__execute(f'get value of attribute "{attribute}"')
        except AppleScriptError:
            return None

    def __execute(self, command, params=''):
        """Execute a command.
               :param str command: The name of the command to _execute as a string.
               :param str params: Command parameters.
               :return str: Execution output."""
        _command = f'{command} {params}' if params else f'{command}'
        return as_wrapper.tell_app_process(command=_command, app_process=self.name)

    hidden = property(is_hidden, set_hidden)
    frontmost = property(is_frontmost, set_frontmost)


class Finder(Application):
    def __init__(self):
        super().__init__('Finder')

    @property
    def window_path(self) -> str:
        return as_wrapper.tell_app(self.name, 'return POSIX path of (target of last Finder window as alias)')

    def eject_mounted_disks(self) -> None:
        return as_wrapper.tell_app(self.name, 'eject (every disk whose ejectable is true)')

    def move_file_to_trash(self, target: str) -> None:
        self.show_destination(target)
        assert self.is_frontmost
        time.sleep(5)
        as_wrapper.send_keycode(as_wrapper.key_codes['delete'], 'command')
        time.sleep(3)
        self.close_windows()

    def show_destination(self, destination: str) -> None:
        if not os.path.exists(destination):
            raise FileNotFoundError(destination)
        self.activate()
        as_wrapper.tell_app(self.name, f'reveal POSIX file "{destination}"')

    def open_destination(self, destination: str) -> None:
        if not os.path.exists(destination):
            raise FileNotFoundError(destination)
        self.activate()
        as_wrapper.tell_app(self.name, f'open POSIX file "{destination}"')


class SystemPreferences(Application):
    anchors: Dict[str, str] = {'full_disk_access': 'Privacy_AllFiles', 'camera_access': 'Privacy_Camera'}

    def __init__(self):
        super().__init__('System Preferences')

    @property
    def current_pane_id(self):
        return as_wrapper.tell_app(self.name, f'return id of current pane')

    def authorize(self):
        return as_wrapper.tell_app(self.name, f'tell current pane to authorize')

    def show_anchor(self, anchor: str, pane_id='com.apple.preference.security'):
        return as_wrapper.tell_app(self.name, f'reveal anchor "{anchor}" of pane "{pane_id}"')


class Installer(Application):
    def __init__(self):
        super().__init__('Installer')

    def open_destination(self, destination: str) -> None:
        destination = destination.strip()
        if not os.path.exists(destination):
            raise FileNotFoundError(destination)
        macos.shell_executor.execute(f'open "{destination}" -a Installer')
        assert self.did_launch
        assert self.window.is_visible
        self.frontmost = True


finder = Finder()
system_preferences_app = SystemPreferences()



import fnmatch
from dataclasses import dataclass
from typing import List

import AppKit
from ApplicationServices import (AXIsProcessTrusted, AXUIElementCreateApplication, AXUIElementCreateSystemWide, CFEqual)
from PyObjCTools import AppHelper

from macuitest.lib.elements.native.callbacks import (get_element_action_names, get_element_attribute_names,
                                                     get_accessibility_element_attribute, get_accessibility_object_pid,
                                                     check_attribute_settable, perform_action_on_element,
                                                     set_attribute_value, set_accessibility_api_timeout,
                                                     get_accessibility_object_on_screen_position, )
from macuitest.lib.elements.native.converter import Converter
from macuitest.lib.elements.native.errors import (AXError, AXErrorAPIDisabled, AXErrorCannotComplete,
                                                  AXErrorIllegalArgument, AXErrorNotImplemented, AXErrorNoValue,
                                                  AXErrorUnsupported, )
from macuitest.lib.elements.native.element_lookup import LookUpMixin


@dataclass
class NSApplicationActivationOptions:
    Default: int = 1
    ActivateAllWindows: int = 2
    ActivateIgnoringOtherWindows: int = 3


def is_accessibility_enabled():
    """Return the status of accessibility on the system."""
    return AXIsProcessTrusted()


def get_frontmost_pid():
    """Return the process ID of the application in the foreground"""
    return AppKit.NSWorkspace.sharedWorkspace().frontmostApplication().processIdentifier()


def get_running_apps():
    """Get a list of the running applications"""
    AppHelper.callLater(1, AppHelper.stopEventLoop)
    AppHelper.runConsoleEventLoop()
    work_space = AppKit.NSWorkspace.sharedWorkspace()
    return work_space.runningApplications()


def launch_app_by_bundle_id(bundle_id: str):
    work_space = AppKit.NSWorkspace.sharedWorkspace()
    response = work_space.launchAppWithBundleIdentifier_options_additionalEventParamDescriptor_launchIdentifier_(
        bundle_id,
        AppKit.NSWorkspaceLaunchAllowingClassicStartup,
        AppKit.NSAppleEventDescriptor.nullDescriptor(),
        None,
    )
    if not response[0]:
        raise RuntimeError(f'Error launching specified application. {response}')


def launch_app_by_bundle_path(bundle_path: str, arguments=None):
    if arguments is None:
        arguments = []
    bundle_url = AppKit.NSURL.fileURLWithPath_(bundle_path)
    workspace = AppKit.NSWorkspace.sharedWorkspace()
    configuration = {AppKit.NSWorkspaceLaunchConfigurationArguments: arguments}
    return workspace.launchApplicationAtURL_options_configuration_error_(
        bundle_url,
        AppKit.NSWorkspaceLaunchAllowingClassicStartup,
        configuration, None
    )


def terminate_app_by_bundle_id(bundle_id: str):
    apps = _running_apps_with_bundle_id(bundle_id)
    if apps:
        apps[0].terminate()


def _running_apps_with_bundle_id(bundle_id: str):
    """Return an array of NSRunningApplications."""
    return AppKit.NSRunningApplication.runningApplicationsWithBundleIdentifier_(bundle_id)


class AXUIElement:
    def __init__(self, ref=None):
        self.ref = ref
        self.converter = Converter(self.__class__)

    def __repr__(self):
        _class_ = self.__class__.__name__
        for identifier in ('AXTitle', 'AXValue', 'AXRoleDescription'):
            title = getattr(self, identifier, None)
            if title:
                break
        return f'<{_class_} {getattr(self, "AXRole", "<No role!>")} "{title}">'

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.ref is None and other.ref is None:
            return True
        if self.ref is None or other.ref is None:
            return False
        return CFEqual(self.ref, other.ref)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattr__(self, item):
        if item in self.ax_attributes:
            return self._get_ax_attribute(item)
        elif item in self.ax_actions:
            def perform_ax_action():
                self._perform_ax_action(item)

            return perform_ax_action
        else:
            raise AttributeError(f'Specified object has no attribute: "{item}"')

    def __setattr__(self, key, value):
        if key.startswith('AX'):
            try:
                if key in self.ax_attributes:
                    self._set_ax_attribute(key, value)
            except AXErrorIllegalArgument:
                pass
        else:
            super(AXUIElement, self).__setattr__(key, value)

    def __dir__(self):
        return self.ax_attributes + self.ax_actions + list(self.__dict__.keys()) + dir(super(AXUIElement, self))

    @classmethod
    def from_bundle_id(cls, bundle_id):
        """Get application by the specified bundle ID."""
        matches = _running_apps_with_bundle_id(bundle_id)
        if not matches:
            raise ValueError(f'"{bundle_id}" not found among running apps.')
        return cls.from_pid(matches[0].processIdentifier())

    @classmethod
    def from_localized_name(cls, name):
        """Get the application by the specified localized name. Wildcards are also allowed."""
        for app in get_running_apps():
            if fnmatch.fnmatch(app.localizedName(), name):
                pid = app.processIdentifier()
                return cls.from_pid(pid)
        raise ValueError(f'"{name}" not found among running applications.')

    @classmethod
    def from_frontmost(cls):
        """Create an instance with the AXUIElementRef for the frontmost application."""
        for app in get_running_apps():
            pid = app.processIdentifier()
            ref = cls.from_pid(pid)
            try:
                if ref.AXFrontmost:
                    return ref
            except (AttributeError, AXErrorUnsupported, AXErrorCannotComplete, AXErrorNotImplemented):
                # Some applications do not have an explicit GUI and so will not have an AXFrontmost attribute.
                pass
        raise ValueError('No GUI applications found.')

    @classmethod
    def from_any_app_with_window(cls):
        """Create an instance with the AXUIElementRef for a random application that has windows."""
        for app in get_running_apps():
            pid = app.processIdentifier()
            ref = cls.from_pid(pid)
            if hasattr(ref, 'windows') and len(ref.windows()) > 0:
                return ref
        raise ValueError('No GUI applications found.')

    @classmethod
    def from_pid(cls, pid: int):
        """Get application by the specified process ID."""
        return cls(ref=AXUIElementCreateApplication(pid))

    @classmethod
    def systemwide(cls):
        """Create an instance with the AXUIElementRef for the system-wide accessibility object."""
        return cls(ref=AXUIElementCreateSystemWide())

    @property
    def ax_actions(self) -> List[str]:
        """Get the list of actions available on the AXUIElement."""
        try:
            return list(get_element_action_names(self.ref))
        except AXError:
            return []

    @property
    def ax_attributes(self) -> List[str]:
        """Get the list of attributes available on the AXUIElement."""
        try:
            return list(get_element_attribute_names(self.ref))
        except AXError:
            return []

    @property
    def bundle_id(self) -> str:
        """Get the AXUIElement's bundle identifier"""
        return self._running_app.bundleIdentifier()

    @property
    def pid(self) -> int:
        """Get the AXUIElement's process ID"""
        return get_accessibility_object_pid(self.ref)

    @property
    def _running_app(self):
        running_application = AppKit.NSRunningApplication
        return running_application.runningApplicationWithProcessIdentifier_(self.pid)

    def set_timeout(self, timeout: int) -> None:
        """Set the timeout value used in the accessibility API."""
        set_accessibility_api_timeout(self.ref, timeout)

    def activate(self) -> bool:
        """Activate the application."""
        return self._running_app.activateWithOptions_(NSApplicationActivationOptions.ActivateIgnoringOtherWindows)

    def _get_ax_attribute(self, item: str):
        """Get the value of the the specified attribute."""
        if item in self.ax_attributes:
            try:
                attribute_value = get_accessibility_element_attribute(self.ref, item)
                return self.converter.convert_value(attribute_value)
            except AXErrorNoValue:
                if item == 'AXChildren':
                    return []
                return None

        raise AttributeError(f'"{type(self)}" object has no attribute "{item}"')

    def _set_ax_attribute(self, name, value):
        """Set the specified attribute to the specified value."""
        is_settable = check_attribute_settable(self.ref, name)
        if not is_settable:
            raise AXErrorUnsupported('Attribute is not settable')
        set_attribute_value(self.ref, name, value)

    def _perform_ax_action(self, name):
        """Perform specified action on the element."""
        perform_action_on_element(self.ref, name)


class NativeUIElement(AXUIElement, LookUpMixin):
    """Expose the accessibility API in the simplest, most natural way possible."""

    def __init__(self, ref=None):
        super(NativeUIElement, self).__init__(ref=ref)
        self.__application = None

    def set_string(self, attribute, string: str):
        """Set the specified attribute to the specified string."""
        return self.__setattr__(attribute, string)

    def get_element_at_position(self, x: int, y: int):
        """Return the AXUIElement at the given coordinates."""
        if self.ref is None:
            raise AXErrorUnsupported('Operation not supported on null element references')
        element = get_accessibility_object_on_screen_position(self.ref, x, y)
        return self.__class__(element)

    def get_menu_item(self, *args):
        """ Return the specified menu item.
            Example - refer to items by name:
                app.get_menu_item('File', 'New').Press()
                app.get_menu_item('Edit', 'Insert', 'Line Break').Press()
            Refer to items by index:
                app.get_menu_item(1, 0).Press()
            Refer to items by mix-n-match:
                app.get_menu_item(1, 'About TextEdit').Press()
        """
        return self.menu_items(self.application.AXMenuBar, *args)

    @property
    def localized_name(self):
        """Return the localized name of the application."""
        return self.application.AXTitle

    @property
    def application(self):
        """Get the base application UIElement.
                If the UIElement is a child of the application, it will try
                to get the AXParent until it reaches the top application level element.
        """
        if self.__application:
            return self.__application
        app = self
        while 'AXParent' in app.ax_attributes:
            app = app.AXParent
        self.__application = app
        return self.__application



typeBoolean = b'bool'
typeChar = b'TEXT'
typeStyledUnicodeText = b'sutx'
typeEncodedString = b'encs'
typeUnicodeText = b'utxt'
typeCString = b'cstr'
typePString = b'pstr'

typeUTF16ExternalRepresentation = b'ut16'
typeUTF8Text = b'utf8'

typeSInt16 = b'shor'
typeUInt16 = b'ushr'
typeSInt32 = b'long'
typeUInt32 = b'magn'
typeSInt64 = b'comp'
typeUInt64 = b'ucom'
typeIEEE32BitFloatingPoint = b'sing'
typeIEEE64BitFloatingPoint = b'doub'
type128BitFloatingPoint = b'ldbl'
typeDecimalStruct = b'decm'

typeSMInt = typeSInt16
typeShortInteger = typeSInt16
typeInteger = typeSInt32
typeLongInteger = typeSInt32
typeMagnitude = typeUInt32
typeComp = typeSInt64
typeSMFloat = typeIEEE32BitFloatingPoint
typeShortFloat = typeIEEE32BitFloatingPoint
typeFloat = typeIEEE64BitFloatingPoint
typeLongFloat = typeIEEE64BitFloatingPoint
typeExtended = b'exte'

typeAEList = b'list'
typeAERecord = b'reco'
typeAppleEvent = b'aevt'
typeEventRecord = b'evrc'
typeTrue = b'true'
typeFalse = b'fals'
typeAlias = b'alis'
typeEnumerated = b'enum'
typeType = b'type'
typeAppParameters = b'appa'
typeProperty = b'prop'
typeFSRef = b'fsrf'
typeFileURL = b'furl'
typeKeyword = b'keyw'
typeSectionH = b'sect'
typeWildCard = b'****'
typeApplSignature = b'sign'
typeQDRectangle = b'qdrt'
typeFixed = b'fixd'
typeProcessSerialNumber = b'psn '
typeApplicationURL = b'aprl'
typeNull = b'null'

typeFSS = b'fss '

typeCFAttributedStringRef = b'cfas'
typeCFMutableAttributedStringRef = b'cfaa'
typeCFStringRef = b'cfst'
typeCFMutableStringRef = b'cfms'
typeCFArrayRef = b'cfar'
typeCFMutableArrayRef = b'cfma'
typeCFDictionaryRef = b'cfdc'
typeCFMutableDictionaryRef = b'cfmd'
typeCFNumberRef = b'cfnb'
typeCFBooleanRef = b'cftf'
typeCFTypeRef = b'cfty'

typeKernelProcessID = b'kpid'
typeMachPort = b'port'

typeApplicationBundleID = b'bund'

keyTransactionIDAttr = b'tran'
keyReturnIDAttr = b'rtid'
keyEventClassAttr = b'evcl'
keyEventIDAttr = b'evid'
keyAddressAttr = b'addr'
keyOptionalKeywordAttr = b'optk'
keyTimeoutAttr = b'timo'
keyInteractLevelAttr = b'inte'
keyEventSourceAttr = b'esrc'
keyMissedKeywordAttr = b'miss'
keyOriginalAddressAttr = b'from'
keyAcceptTimeoutAttr = b'actm'
keyReplyRequestedAttr = b'repq'

kAEDebugPOSTHeader = (1 << 0)
kAEDebugReplyHeader = (1 << 1)
kAEDebugXMLRequest = (1 << 2)
kAEDebugXMLResponse = (1 << 3)
kAEDebugXMLDebugAll = 0xFFFFFFFF

kSOAP1999Schema = b'ss99'
kSOAP2001Schema = b'ss01'

keyUserNameAttr = b'unam'
keyUserPasswordAttr = b'pass'
keyDisableAuthenticationAttr = b'auth'
keyXMLDebuggingAttr = b'xdbg'
kAERPCClass = b'rpc '
kAEXMLRPCScheme = b'RPC2'
kAESOAPScheme = b'SOAP'
kAESharedScriptHandler = b'wscp'
keyRPCMethodName = b'meth'
keyRPCMethodParam = b'parm'
keyRPCMethodParamOrder = b'/ord'
keyAEPOSTHeaderData = b'phed'
keyAEReplyHeaderData = b'rhed'
keyAEXMLRequestData = b'xreq'
keyAEXMLReplyData = b'xrep'
keyAdditionalHTTPHeaders = b'ahed'
keySOAPAction = b'sact'
keySOAPMethodNameSpace = b'mspc'
keySOAPMethodNameSpaceURI = b'mspu'
keySOAPSchemaVersion = b'ssch'

keySOAPStructureMetaData = b'/smd'
keySOAPSMDNamespace = b'ssns'
keySOAPSMDNamespaceURI = b'ssnu'
keySOAPSMDType = b'sstp'

kAEUseHTTPProxyAttr = b'xupr'
kAEHTTPProxyPortAttr = b'xhtp'
kAEHTTPProxyHostAttr = b'xhth'

kAESocks4Protocol = 4
kAESocks5Protocol = 5

kAEUseSocksAttr = b'xscs'
kAESocksProxyAttr = b'xsok'
kAESocksHostAttr = b'xshs'
kAESocksPortAttr = b'xshp'
kAESocksUserAttr = b'xshu'
kAESocksPasswordAttr = b'xshw'

kAEDescListFactorNone = 0
kAEDescListFactorType = 4
kAEDescListFactorTypeAndSize = 8

kAutoGenerateReturnID = -1
kAnyTransactionID = 0

kAEDataArray = 0
kAEPackedArray = 1
kAEDescArray = 3
kAEKeyDescArray = 4

kAEHandleArray = 2

kAENormalPriority = 0x00000000
kAEHighPriority = 0x00000001

kAENoReply = 0x00000001
kAEQueueReply = 0x00000002
kAEWaitReply = 0x00000003
kAEDontReconnect = 0x00000080
kAEWantReceipt = 0x00000200
kAENeverInteract = 0x00000010
kAECanInteract = 0x00000020
kAEAlwaysInteract = 0x00000030
kAECanSwitchLayer = 0x00000040
kAEDontRecord = 0x00001000
kAEDontExecute = 0x00002000
kAEProcessNonReplyEvents = 0x00008000

kAEDefaultTimeout = -1
kNoTimeOut = -2

# AEHelpers.h

aeBuildSyntaxNoErr = 0
aeBuildSyntaxBadToken = 1
aeBuildSyntaxBadEOF = 2
aeBuildSyntaxNoEOF = 3
aeBuildSyntaxBadNegative = 4
aeBuildSyntaxMissingQuote = 5
aeBuildSyntaxBadHex = 6
aeBuildSyntaxOddHex = 7
aeBuildSyntaxNoCloseHex = 8
aeBuildSyntaxUncoercedHex = 9
aeBuildSyntaxNoCloseString = 10
aeBuildSyntaxBadDesc = 11
aeBuildSyntaxBadData = 12
aeBuildSyntaxNoCloseParen = 13
aeBuildSyntaxNoCloseBracket = 14
aeBuildSyntaxNoCloseBrace = 15
aeBuildSyntaxNoKey = 16
aeBuildSyntaxNoColon = 17
aeBuildSyntaxCoercedList = 18
aeBuildSyntaxUncoercedDoubleAt = 19

# AEMach.h

keyReplyPortAttr = b'repp'

typeReplyPortAttr = keyReplyPortAttr

# AEObjects.h

kAEAND = b'AND '
kAEOR = b'OR  '
kAENOT = b'NOT '
kAEFirst = b'firs'
kAELast = b'last'
kAEMiddle = b'midd'
kAEAny = b'any '
kAEAll = b'all '
kAENext = b'next'
kAEPrevious = b'prev'
keyAECompOperator = b'relo'
keyAELogicalTerms = b'term'
keyAELogicalOperator = b'logc'
keyAEObject1 = b'obj1'
keyAEObject2 = b'obj2'
keyAEDesiredClass = b'want'
keyAEContainer = b'from'
keyAEKeyForm = b'form'
keyAEKeyData = b'seld'

keyAERangeStart = b'star'
keyAERangeStop = b'stop'
keyDisposeTokenProc = b'xtok'
keyAECompareProc = b'cmpr'
keyAECountProc = b'cont'
keyAEMarkTokenProc = b'mkid'
keyAEMarkProc = b'mark'
keyAEAdjustMarksProc = b'adjm'
keyAEGetErrDescProc = b'indc'

formAbsolutePosition = b'indx'
formRelativePosition = b'rele'
formTest = b'test'
formRange = b'rang'
formPropertyID = b'prop'
formName = b'name'
formUniqueID = b'ID  '
typeObjectSpecifier = b'obj '
typeObjectBeingExamined = b'exmn'
typeCurrentContainer = b'ccnt'
typeToken = b'toke'
typeRelativeDescriptor = b'rel '
typeAbsoluteOrdinal = b'abso'
typeIndexDescriptor = b'inde'
typeRangeDescriptor = b'rang'
typeLogicalDescriptor = b'logi'
typeCompDescriptor = b'cmpd'
typeOSLTokenList = b'ostl'

kAEIDoMinimum = 0x0000
kAEIDoWhose = 0x0001
kAEIDoMarking = 0x0004
kAEPassSubDescs = 0x0008
kAEResolveNestedLists = 0x0010
kAEHandleSimpleRanges = 0x0020
kAEUseRelativeIterators = 0x0040

typeWhoseDescriptor = b'whos'
formWhose = b'whos'
typeWhoseRange = b'wrng'
keyAEWhoseRangeStart = b'wstr'
keyAEWhoseRangeStop = b'wstp'
keyAEIndex = b'kidx'
keyAETest = b'ktst'

# AEPackObject.h


# AERegistry.h

cAEList = b'list'
cApplication = b'capp'
cArc = b'carc'
cBoolean = b'bool'
cCell = b'ccel'
cChar = b'cha '
cColorTable = b'clrt'
cColumn = b'ccol'
cDocument = b'docu'
cDrawingArea = b'cdrw'
cEnumeration = b'enum'
cFile = b'file'
cFixed = b'fixd'
cFixedPoint = b'fpnt'
cFixedRectangle = b'frct'
cGraphicLine = b'glin'
cGraphicObject = b'cgob'
cGraphicShape = b'cgsh'
cGraphicText = b'cgtx'
cGroupedGraphic = b'cpic'

cInsertionLoc = b'insl'
cInsertionPoint = b'cins'
cIntlText = b'itxt'
cIntlWritingCode = b'intl'
cItem = b'citm'
cLine = b'clin'
cLongDateTime = b'ldt '
cLongFixed = b'lfxd'
cLongFixedPoint = b'lfpt'
cLongFixedRectangle = b'lfrc'
cLongInteger = b'long'
cLongPoint = b'lpnt'
cLongRectangle = b'lrct'
cMachineLoc = b'mLoc'
cMenu = b'cmnu'
cMenuItem = b'cmen'
cObject = b'cobj'
cObjectSpecifier = b'obj '
cOpenableObject = b'coob'
cOval = b'covl'

cParagraph = b'cpar'
cPICT = b'PICT'
cPixel = b'cpxl'
cPixelMap = b'cpix'
cPolygon = b'cpgn'
cProperty = b'prop'
cQDPoint = b'QDpt'
cQDRectangle = b'qdrt'
cRectangle = b'crec'
cRGBColor = b'cRGB'
cRotation = b'trot'
cRoundedRectangle = b'crrc'
cRow = b'crow'
cSelection = b'csel'
cShortInteger = b'shor'
cTable = b'ctbl'
cText = b'ctxt'
cTextFlow = b'cflo'
cTextStyles = b'tsty'
cType = b'type'

cVersion = b'vers'
cWindow = b'cwin'
cWord = b'cwor'
enumArrows = b'arro'
enumJustification = b'just'
enumKeyForm = b'kfrm'
enumPosition = b'posi'
enumProtection = b'prtn'
enumQuality = b'qual'
enumSaveOptions = b'savo'
enumStyle = b'styl'
enumTransferMode = b'tran'
kAEAbout = b'abou'
kAEAfter = b'afte'
kAEAliasSelection = b'sali'
kAEAllCaps = b'alcp'
kAEArrowAtEnd = b'aren'
kAEArrowAtStart = b'arst'
kAEArrowBothEnds = b'arbo'

kAEAsk = b'ask '
kAEBefore = b'befo'
kAEBeginning = b'bgng'
kAEBeginsWith = b'bgwt'
kAEBeginTransaction = b'begi'
kAEBold = b'bold'
kAECaseSensEquals = b'cseq'
kAECentered = b'cent'
kAEChangeView = b'view'
kAEClone = b'clon'
kAEClose = b'clos'
kAECondensed = b'cond'
kAEContains = b'cont'
kAECopy = b'copy'
kAECoreSuite = b'core'
kAECountElements = b'cnte'
kAECreateElement = b'crel'
kAECreatePublisher = b'cpub'
kAECut = b'cut '
kAEDelete = b'delo'

kAEDoObjectsExist = b'doex'
kAEDoScript = b'dosc'
kAEDrag = b'drag'
kAEDuplicateSelection = b'sdup'
kAEEditGraphic = b'edit'
kAEEmptyTrash = b'empt'
kAEEnd = b'end '
kAEEndsWith = b'ends'
kAEEndTransaction = b'endt'
kAEEquals = b'=   '
kAEExpanded = b'pexp'
kAEFast = b'fast'
kAEFinderEvents = b'FNDR'
kAEFormulaProtect = b'fpro'
kAEFullyJustified = b'full'
kAEGetClassInfo = b'qobj'
kAEGetData = b'getd'
kAEGetDataSize = b'dsiz'
kAEGetEventInfo = b'gtei'
kAEGetInfoSelection = b'sinf'

kAEGetPrivilegeSelection = b'sprv'
kAEGetSuiteInfo = b'gtsi'
kAEGreaterThan = b'>   '
kAEGreaterThanEquals = b'>=  '
kAEGrow = b'grow'
kAEHidden = b'hidn'
kAEHiQuality = b'hiqu'
kAEImageGraphic = b'imgr'
kAEIsUniform = b'isun'
kAEItalic = b'ital'
kAELeftJustified = b'left'
kAELessThan = b'<   '
kAELessThanEquals = b'<=  '
kAELowercase = b'lowc'
kAEMakeObjectsVisible = b'mvis'
kAEMiscStandards = b'misc'
kAEModifiable = b'modf'
kAEMove = b'move'
kAENo = b'no  '
kAENoArrow = b'arno'

kAENonmodifiable = b'nmod'
kAEOpen = b'odoc'
kAEOpenSelection = b'sope'
kAEOutline = b'outl'
kAEPageSetup = b'pgsu'
kAEPaste = b'past'
kAEPlain = b'plan'
kAEPrint = b'pdoc'
kAEPrintSelection = b'spri'
kAEPrintWindow = b'pwin'
kAEPutAwaySelection = b'sput'
kAEQDAddOver = b'addo'
kAEQDAddPin = b'addp'
kAEQDAdMax = b'admx'
kAEQDAdMin = b'admn'
kAEQDBic = b'bic '
kAEQDBlend = b'blnd'
kAEQDCopy = b'cpy '
kAEQDNotBic = b'nbic'
kAEQDNotCopy = b'ncpy'

kAEQDNotOr = b'ntor'
kAEQDNotXor = b'nxor'
kAEQDOr = b'or  '
kAEQDSubOver = b'subo'
kAEQDSubPin = b'subp'
kAEQDSupplementalSuite = b'qdsp'
kAEQDXor = b'xor '
kAEQuickdrawSuite = b'qdrw'
kAEQuitAll = b'quia'
kAERedo = b'redo'
kAERegular = b'regl'
kAEReopenApplication = b'rapp'
kAEReplace = b'rplc'
kAERequiredSuite = b'reqd'
kAERestart = b'rest'
kAERevealSelection = b'srev'
kAERevert = b'rvrt'
kAERightJustified = b'rght'
kAESave = b'save'
kAESelect = b'slct'
kAESetData = b'setd'

kAESetPosition = b'posn'
kAEShadow = b'shad'
kAEShowClipboard = b'shcl'
kAEShutDown = b'shut'
kAESleep = b'slep'
kAESmallCaps = b'smcp'
kAESpecialClassProperties = b'c@#!'
kAEStrikethrough = b'strk'
kAESubscript = b'sbsc'
kAESuperscript = b'spsc'
kAETableSuite = b'tbls'
kAETextSuite = b'TEXT'
kAETransactionTerminated = b'ttrm'
kAEUnderline = b'undl'
kAEUndo = b'undo'
kAEWholeWordEquals = b'wweq'
kAEYes = b'yes '
kAEZoom = b'zoom'

kAELogOut = b'logo'
kAEReallyLogOut = b'rlgo'
kAEShowRestartDialog = b'rrst'
kAEShowShutdownDialog = b'rsdn'

kAEMouseClass = b'mous'
kAEDown = b'down'
kAEUp = b'up  '
kAEMoved = b'move'
kAEStoppedMoving = b'stop'
kAEWindowClass = b'wind'
kAEUpdate = b'updt'
kAEActivate = b'actv'
kAEDeactivate = b'dact'
kAECommandClass = b'cmnd'
kAEKeyClass = b'keyc'
kAERawKey = b'rkey'
kAEVirtualKey = b'keyc'
kAENavigationKey = b'nave'
kAEAutoDown = b'auto'
kAEApplicationClass = b'appl'
kAESuspend = b'susp'
kAEResume = b'rsme'
kAEDiskEvent = b'disk'
kAENullEvent = b'null'
kAEWakeUpEvent = b'wake'
kAEScrapEvent = b'scrp'
kAEHighLevel = b'high'

keyAEAngle = b'kang'
keyAEArcAngle = b'parc'

keyAEBaseAddr = b'badd'
keyAEBestType = b'pbst'
keyAEBgndColor = b'kbcl'
keyAEBgndPattern = b'kbpt'
keyAEBounds = b'pbnd'
keyAECellList = b'kclt'
keyAEClassID = b'clID'
keyAEColor = b'colr'
keyAEColorTable = b'cltb'
keyAECurveHeight = b'kchd'
keyAECurveWidth = b'kcwd'
keyAEDashStyle = b'pdst'
keyAEData = b'data'
keyAEDefaultType = b'deft'
keyAEDefinitionRect = b'pdrt'
keyAEDescType = b'dstp'
keyAEDestination = b'dest'
keyAEDoAntiAlias = b'anta'
keyAEDoDithered = b'gdit'
keyAEDoRotate = b'kdrt'

keyAEDoScale = b'ksca'
keyAEDoTranslate = b'ktra'
keyAEEditionFileLoc = b'eloc'
keyAEElements = b'elms'
keyAEEndPoint = b'pend'
keyAEEventClass = b'evcl'
keyAEEventID = b'evti'
keyAEFile = b'kfil'
keyAEFileType = b'fltp'
keyAEFillColor = b'flcl'
keyAEFillPattern = b'flpt'
keyAEFlipHorizontal = b'kfho'
keyAEFlipVertical = b'kfvt'
keyAEFont = b'font'
keyAEFormula = b'pfor'
keyAEGraphicObjects = b'gobs'
keyAEID = b'ID  '
keyAEImageQuality = b'gqua'
keyAEInsertHere = b'insh'
keyAEKeyForms = b'keyf'

keyAEKeyword = b'kywd'
keyAELevel = b'levl'
keyAELineArrow = b'arro'
keyAEName = b'pnam'
keyAENewElementLoc = b'pnel'
keyAEObject = b'kobj'
keyAEObjectClass = b'kocl'
keyAEOffStyles = b'ofst'
keyAEOnStyles = b'onst'
keyAEParameters = b'prms'
keyAEParamFlags = b'pmfg'
keyAEPenColor = b'ppcl'
keyAEPenPattern = b'pppa'
keyAEPenWidth = b'ppwd'
keyAEPixelDepth = b'pdpt'
keyAEPixMapMinus = b'kpmm'
keyAEPMTable = b'kpmt'
keyAEPointList = b'ptlt'
keyAEPointSize = b'ptsz'
keyAEPosition = b'kpos'

keyAEPropData = b'prdt'
keyAEProperties = b'qpro'
keyAEProperty = b'kprp'
keyAEPropFlags = b'prfg'
keyAEPropID = b'prop'
keyAEProtection = b'ppro'
keyAERenderAs = b'kren'
keyAERequestedType = b'rtyp'
keyAEResult = b'----'
keyAEResultInfo = b'rsin'
keyAERotation = b'prot'
keyAERotPoint = b'krtp'
keyAERowList = b'krls'
keyAESaveOptions = b'savo'
keyAEScale = b'pscl'
keyAEScriptTag = b'psct'
keyAESearchText = b'stxt'
keyAEShowWhere = b'show'
keyAEStartAngle = b'pang'
keyAEStartPoint = b'pstp'
keyAEStyles = b'ksty'

keyAESuiteID = b'suit'
keyAEText = b'ktxt'
keyAETextColor = b'ptxc'
keyAETextFont = b'ptxf'
keyAETextPointSize = b'ptps'
keyAETextStyles = b'txst'
keyAETextLineHeight = b'ktlh'
keyAETextLineAscent = b'ktas'
keyAETheText = b'thtx'
keyAETransferMode = b'pptm'
keyAETranslation = b'ptrs'
keyAETryAsStructGraf = b'toog'
keyAEUniformStyles = b'ustl'
keyAEUpdateOn = b'pupd'
keyAEUserTerm = b'utrm'
keyAEWindow = b'wndw'
keyAEWritingCode = b'wrcd'

keyMiscellaneous = b'fmsc'
keySelection = b'fsel'
keyWindow = b'kwnd'
keyWhen = b'when'
keyWhere = b'wher'
keyModifiers = b'mods'
keyKey = b'key '
keyKeyCode = b'code'
keyKeyboard = b'keyb'
keyDriveNumber = b'drv#'
keyErrorCode = b'err#'
keyHighLevelClass = b'hcls'
keyHighLevelID = b'hid '

pArcAngle = b'parc'
pBackgroundColor = b'pbcl'
pBackgroundPattern = b'pbpt'
pBestType = b'pbst'
pBounds = b'pbnd'
pClass = b'pcls'
pClipboard = b'pcli'
pColor = b'colr'
pColorTable = b'cltb'
pContents = b'pcnt'
pCornerCurveHeight = b'pchd'
pCornerCurveWidth = b'pcwd'
pDashStyle = b'pdst'
pDefaultType = b'deft'
pDefinitionRect = b'pdrt'
pEnabled = b'enbl'
pEndPoint = b'pend'
pFillColor = b'flcl'
pFillPattern = b'flpt'
pFont = b'font'

pFormula = b'pfor'
pGraphicObjects = b'gobs'
pHasCloseBox = b'hclb'
pHasTitleBar = b'ptit'
pID = b'ID  '
pIndex = b'pidx'
pInsertionLoc = b'pins'
pIsFloating = b'isfl'
pIsFrontProcess = b'pisf'
pIsModal = b'pmod'
pIsModified = b'imod'
pIsResizable = b'prsz'
pIsStationeryPad = b'pspd'
pIsZoomable = b'iszm'
pIsZoomed = b'pzum'
pItemNumber = b'itmn'
pJustification = b'pjst'
pLineArrow = b'arro'
pMenuID = b'mnid'
pName = b'pnam'

pNewElementLoc = b'pnel'
pPenColor = b'ppcl'
pPenPattern = b'pppa'
pPenWidth = b'ppwd'
pPixelDepth = b'pdpt'
pPointList = b'ptlt'
pPointSize = b'ptsz'
pProtection = b'ppro'
pRotation = b'prot'
pScale = b'pscl'
pScript = b'scpt'
pScriptTag = b'psct'
pSelected = b'selc'
pSelection = b'sele'
pStartAngle = b'pang'
pStartPoint = b'pstp'
pTextColor = b'ptxc'
pTextFont = b'ptxf'
pTextItemDelimiters = b'txdl'
pTextPointSize = b'ptps'

pTextStyles = b'txst'
pTransferMode = b'pptm'
pTranslation = b'ptrs'
pUniformStyles = b'ustl'
pUpdateOn = b'pupd'
pUserSelection = b'pusl'
pVersion = b'vers'
pVisible = b'pvis'

typeAEText = b'tTXT'
typeArc = b'carc'
typeBest = b'best'
typeCell = b'ccel'
typeClassInfo = b'gcli'
typeColorTable = b'clrt'
typeColumn = b'ccol'
typeDashStyle = b'tdas'
typeData = b'tdta'
typeDrawingArea = b'cdrw'
typeElemInfo = b'elin'
typeEnumeration = b'enum'
typeEPS = b'EPS '
typeEventInfo = b'evin'

typeFinderWindow = b'fwin'
typeFixedPoint = b'fpnt'
typeFixedRectangle = b'frct'
typeGraphicLine = b'glin'
typeGraphicText = b'cgtx'
typeGroupedGraphic = b'cpic'
typeInsertionLoc = b'insl'
typeIntlText = b'itxt'
typeIntlWritingCode = b'intl'
typeLongDateTime = b'ldt '
typeCFAbsoluteTime = b'cfat'
typeISO8601DateTime = b'isot'
typeLongFixed = b'lfxd'
typeLongFixedPoint = b'lfpt'
typeLongFixedRectangle = b'lfrc'
typeLongPoint = b'lpnt'
typeLongRectangle = b'lrct'
typeMachineLoc = b'mLoc'
typeOval = b'covl'
typeParamInfo = b'pmin'
typePict = b'PICT'

typePixelMap = b'cpix'
typePixMapMinus = b'tpmm'
typePolygon = b'cpgn'
typePropInfo = b'pinf'
typePtr = b'ptr '
typeQDPoint = b'QDpt'
typeQDRegion = b'Qrgn'
typeRectangle = b'crec'
typeRGB16 = b'tr16'
typeRGB96 = b'tr96'
typeRGBColor = b'cRGB'
typeRotation = b'trot'
typeRoundedRectangle = b'crrc'
typeRow = b'crow'
typeScrapStyles = b'styl'
typeScript = b'scpt'
typeStyledText = b'STXT'
typeSuiteInfo = b'suin'
typeTable = b'ctbl'
typeTextStyles = b'tsty'

typeTIFF = b'TIFF'
typeJPEG = b'JPEG'
typeGIF = b'GIFf'
typeVersion = b'vers'

kAEMenuClass = b'menu'
kAEMenuSelect = b'mhit'
kAEMouseDown = b'mdwn'
kAEMouseDownInBack = b'mdbk'
kAEKeyDown = b'kdwn'
kAEResized = b'rsiz'
kAEPromise = b'prom'

keyMenuID = b'mid '
keyMenuItem = b'mitm'
keyCloseAllWindows = b'caw '
keyOriginalBounds = b'obnd'
keyNewBounds = b'nbnd'
keyLocalWhere = b'lwhr'

typeHIMenu = b'mobj'
typeHIWindow = b'wobj'

kBySmallIcon = 0
kByIconView = 1
kByNameView = 2
kByDateView = 3
kBySizeView = 4
kByKindView = 5
kByCommentView = 6
kByLabelView = 7
kByVersionView = 8

kAEInfo = 11
kAEMain = 0
kAESharing = 13

kAEZoomIn = 7
kAEZoomOut = 8

kTextServiceClass = b'tsvc'
kUpdateActiveInputArea = b'updt'
kShowHideInputWindow = b'shiw'
kPos2Offset = b'p2st'
kOffset2Pos = b'st2p'
kUnicodeNotFromInputMethod = b'unim'
kGetSelectedText = b'gtxt'
keyAETSMDocumentRefcon = b'refc'
keyAEServerInstance = b'srvi'
keyAETheData = b'kdat'
keyAEFixLength = b'fixl'
keyAEUpdateRange = b'udng'
keyAECurrentPoint = b'cpos'
keyAEBufferSize = b'buff'
keyAEMoveView = b'mvvw'
keyAENextBody = b'nxbd'
keyAETSMScriptTag = b'sclg'
keyAETSMTextFont = b'ktxf'
keyAETSMTextFMFont = b'ktxm'
keyAETSMTextPointSize = b'ktps'
keyAETSMEventRecord = b'tevt'
keyAETSMEventRef = b'tevr'
keyAETextServiceEncoding = b'tsen'
keyAETextServiceMacEncoding = b'tmen'
keyAETSMGlyphInfoArray = b'tgia'
typeTextRange = b'txrn'
typeComponentInstance = b'cmpi'
typeOffsetArray = b'ofay'
typeTextRangeArray = b'tray'
typeLowLevelEventRecord = b'evtr'
typeGlyphInfoArray = b'glia'
typeEventRef = b'evrf'
typeText = typeChar

kTSMOutsideOfBody = 1
kTSMInsideOfBody = 2
kTSMInsideOfActiveInputArea = 3

kNextBody = 1
kPreviousBody = 2

kTSMHiliteCaretPosition = 1
kTSMHiliteRawText = 2
kTSMHiliteSelectedRawText = 3
kTSMHiliteConvertedText = 4
kTSMHiliteSelectedConvertedText = 5
kTSMHiliteBlockFillText = 6
kTSMHiliteOutlineText = 7
kTSMHiliteSelectedText = 8
kTSMHiliteNoHilite = 9

kCaretPosition = kTSMHiliteCaretPosition
kRawText = kTSMHiliteRawText
kSelectedRawText = kTSMHiliteSelectedRawText
kConvertedText = kTSMHiliteConvertedText
kSelectedConvertedText = kTSMHiliteSelectedConvertedText
kBlockFillText = kTSMHiliteBlockFillText
kOutlineText = kTSMHiliteOutlineText
kSelectedText = kTSMHiliteSelectedText

keyAEHiliteRange = b'hrng'
keyAEPinRange = b'pnrg'
keyAEClauseOffsets = b'clau'
keyAEOffset = b'ofst'
keyAEPoint = b'gpos'
keyAELeftSide = b'klef'
keyAERegionClass = b'rgnc'
keyAEDragging = b'bool'

keyAELeadingEdge = keyAELeftSide

typeMeters = b'metr'
typeInches = b'inch'
typeFeet = b'feet'
typeYards = b'yard'
typeMiles = b'mile'
typeKilometers = b'kmtr'
typeCentimeters = b'cmtr'
typeSquareMeters = b'sqrm'
typeSquareFeet = b'sqft'
typeSquareYards = b'sqyd'
typeSquareMiles = b'sqmi'
typeSquareKilometers = b'sqkm'
typeLiters = b'litr'
typeQuarts = b'qrts'
typeGallons = b'galn'
typeCubicMeters = b'cmet'
typeCubicFeet = b'cfet'
typeCubicInches = b'cuin'
typeCubicCentimeter = b'ccmt'
typeCubicYards = b'cyrd'
typeKilograms = b'kgrm'
typeGrams = b'gram'
typeOunces = b'ozs '
typePounds = b'lbs '
typeDegreesC = b'degc'
typeDegreesF = b'degf'
typeDegreesK = b'degk'

kFAServerApp = b'ssrv'
kDoFolderActionEvent = b'fola'
kFolderActionCode = b'actn'
kFolderOpenedEvent = b'fopn'
kFolderClosedEvent = b'fclo'
kFolderWindowMovedEvent = b'fsiz'
kFolderItemsAddedEvent = b'fget'
kFolderItemsRemovedEvent = b'flos'
kItemList = b'flst'
kNewSizeParameter = b'fnsz'
kFASuiteCode = b'faco'
kFAAttachCommand = b'atfa'
kFARemoveCommand = b'rmfa'
kFAEditCommand = b'edfa'
kFAFileParam = b'faal'
kFAIndexParam = b'indx'

kAEInternetSuite = b'gurl'
kAEISWebStarSuite = b'WWW\xBD'

kAEISGetURL = b'gurl'
KAEISHandleCGI = b'sdoc'

cURL = b'url '
cInternetAddress = b'IPAD'
cHTML = b'html'
cFTPItem = b'ftp '

kAEISHTTPSearchArgs = b'kfor'
kAEISPostArgs = b'post'
kAEISMethod = b'meth'
kAEISClientAddress = b'addr'
kAEISUserName = b'user'
kAEISPassword = b'pass'
kAEISFromUser = b'frmu'
kAEISServerName = b'svnm'
kAEISServerPort = b'svpt'
kAEISScriptName = b'scnm'
kAEISContentType = b'ctyp'
kAEISReferrer = b'refr'
kAEISUserAgent = b'Agnt'
kAEISAction = b'Kact'
kAEISActionPath = b'Kapt'
kAEISClientIP = b'Kcip'
kAEISFullRequest = b'Kfrq'

pScheme = b'pusc'
pHost = b'HOST'
pPath = b'FTPc'
pUserName = b'RAun'
pUserPassword = b'RApw'
pDNSForm = b'pDNS'
pURL = b'pURL'
pTextEncoding = b'ptxe'
pFTPKind = b'kind'

eScheme = b'esch'
eurlHTTP = b'http'
eurlHTTPS = b'htps'
eurlFTP = b'ftp '
eurlMail = b'mail'
eurlFile = b'file'
eurlGopher = b'gphr'
eurlTelnet = b'tlnt'
eurlNews = b'news'
eurlSNews = b'snws'
eurlNNTP = b'nntp'
eurlMessage = b'mess'
eurlMailbox = b'mbox'
eurlMulti = b'mult'
eurlLaunch = b'laun'
eurlAFP = b'afp '
eurlAT = b'at  '
eurlEPPC = b'eppc'
eurlRTSP = b'rtsp'
eurlIMAP = b'imap'
eurlNFS = b'unfs'
eurlPOP = b'upop'
eurlLDAP = b'uldp'
eurlUnknown = b'url?'

kConnSuite = b'macc'
cDevSpec = b'cdev'
cAddressSpec = b'cadr'
cADBAddress = b'cadb'
cAppleTalkAddress = b'cat '
cBusAddress = b'cbus'
cEthernetAddress = b'cen '
cFireWireAddress = b'cfw '
cIPAddress = b'cip '
cLocalTalkAddress = b'clt '
cSCSIAddress = b'cscs'
cTokenRingAddress = b'ctok'
cUSBAddress = b'cusb'
pDeviceType = b'pdvt'
pDeviceAddress = b'pdva'
pConduit = b'pcon'
pProtocol = b'pprt'
pATMachine = b'patm'
pATZone = b'patz'
pATType = b'patt'
pDottedDecimal = b'pipd'
pDNS = b'pdns'
pPort = b'ppor'
pNetwork = b'pnet'
pNode = b'pnod'
pSocket = b'psoc'
pSCSIBus = b'pscb'
pSCSILUN = b'pslu'
eDeviceType = b'edvt'
eAddressSpec = b'eads'
eConduit = b'econ'
eProtocol = b'epro'
eADB = b'eadb'
eAnalogAudio = b'epau'
eAppleTalk = b'epat'
eAudioLineIn = b'ecai'
eAudioLineOut = b'ecal'
eAudioOut = b'ecao'
eBus = b'ebus'
eCDROM = b'ecd '
eCommSlot = b'eccm'
eDigitalAudio = b'epda'
eDisplay = b'edds'
eDVD = b'edvd'
eEthernet = b'ecen'
eFireWire = b'ecfw'
eFloppy = b'efd '
eHD = b'ehd '
eInfrared = b'ecir'
eIP = b'epip'
eIrDA = b'epir'
eIRTalk = b'epit'
eKeyboard = b'ekbd'
eLCD = b'edlc'
eLocalTalk = b'eclt'
eMacIP = b'epmi'
eMacVideo = b'epmv'
eMicrophone = b'ecmi'
eModemPort = b'ecmp'
eModemPrinterPort = b'empp'
eModem = b'edmm'
eMonitorOut = b'ecmn'
eMouse = b'emou'
eNuBusCard = b'ednb'
eNuBus = b'enub'
ePCcard = b'ecpc'
ePCIbus = b'ecpi'
ePCIcard = b'edpi'
ePDSslot = b'ecpd'
ePDScard = b'epds'
ePointingDevice = b'edpd'
ePostScript = b'epps'
ePPP = b'eppp'
ePrinterPort = b'ecpp'
ePrinter = b'edpr'
eSvideo = b'epsv'
eSCSI = b'ecsc'
eSerial = b'epsr'
eSpeakers = b'edsp'
eStorageDevice = b'edst'
eSVGA = b'epsg'
eTokenRing = b'etok'
eTrackball = b'etrk'
eTrackpad = b'edtp'
eUSB = b'ecus'
eVideoIn = b'ecvi'
eVideoMonitor = b'edvm'
eVideoOut = b'ecvo'

cKeystroke = b'kprs'
pKeystrokeKey = b'kMsg'
pModifiers = b'kMod'
pKeyKind = b'kknd'
eModifiers = b'eMds'
eOptionDown = b'Kopt'
eCommandDown = b'Kcmd'
eControlDown = b'Kctl'
eShiftDown = b'Ksft'
eCapsLockDown = b'Kclk'
eKeyKind = b'ekst'
eEscapeKey = b'ks5\x00'
eDeleteKey = b'ks3\x00'
eTabKey = b'ks0\x00'
eReturnKey = b'ks\x24\x00'
eClearKey = b'ksG\x00'
eEnterKey = b'ksL\x00'
eUpArrowKey = b'ks\x7E\x00'
eDownArrowKey = b'ks\x7D\x00'
eLeftArrowKey = b'ks\x7B\x00'
eRightArrowKey = b'ks\x7C\x00'
eHelpKey = b'ksr\x00'
eHomeKey = b'kss\x00'
ePageUpKey = b'kst\x00'
ePageDownKey = b'ksy\x00'
eForwardDelKey = b'ksu\x00'
eEndKey = b'ksw\x00'
eF1Key = b'ksz\x00'
eF2Key = b'ksx\x00'
eF3Key = b'ksc\x00'
eF4Key = b'ksv\x00'
eF5Key = b'ks\x60\x00'
eF6Key = b'ksa\x00'
eF7Key = b'ksb\x00'
eF8Key = b'ksd\x00'
eF9Key = b'kse\x00'
eF10Key = b'ksm\x00'
eF11Key = b'ksg\x00'
eF12Key = b'kso\x00'
eF13Key = b'ksi\x00'
eF14Key = b'ksk\x00'
eF15Key = b'ksq\x00'

keyAELaunchedAsLogInItem = b'lgit'
keyAELaunchedAsServiceItem = b'svit'

# AEUserTermTypes.h

kAEUserTerminology = b'aeut'
kAETerminologyExtension = b'aete'
kAEScriptingSizeResource = b'scsz'
kAEOSAXSizeResource = b'osiz'

kAEUTHasReturningParam = 31
kAEUTOptional = 15
kAEUTlistOfItems = 14
kAEUTEnumerated = 13
kAEUTReadWrite = 12
kAEUTChangesState = 12
kAEUTTightBindingFunction = 12
kAEUTEnumsAreTypes = 11
kAEUTEnumListIsExclusive = 10
kAEUTReplyIsReference = 9
kAEUTDirectParamIsReference = 9
kAEUTParamIsReference = 9
kAEUTPropertyIsReference = 9
kAEUTNotDirectParamIsTarget = 8
kAEUTParamIsTarget = 8
kAEUTApostrophe = 3
kAEUTFeminine = 2
kAEUTMasculine = 1
kAEUTPlural = 0

kLaunchToGetTerminology = (1 << 15)
kDontFindAppBySignature = (1 << 14)
kAlwaysSendSubject = (1 << 13)

kReadExtensionTermsMask = (1 << 15)

kOSIZDontOpenResourceFile = 15
kOSIZdontAcceptRemoteEvents = 14
kOSIZOpenWithReadPermission = 13
kOSIZCodeInSharedLibraries = 11

# AppleEvents.h

keyDirectObject = b'----'
keyErrorNumber = b'errn'
keyErrorString = b'errs'
keyProcessSerialNumber = b'psn '
keyPreDispatch = b'phac'
keySelectProc = b'selh'
keyAERecorderCount = b'recr'
keyAEVersion = b'vers'

kCoreEventClass = b'aevt'

kAEOpenApplication = b'oapp'
kAEOpenDocuments = b'odoc'
kAEPrintDocuments = b'pdoc'
kAEOpenContents = b'ocon'
kAEQuitApplication = b'quit'
kAEAnswer = b'ansr'
kAEApplicationDied = b'obit'
kAEShowPreferences = b'pref'

kAEStartRecording = b'reca'
kAEStopRecording = b'recc'
kAENotifyStartRecording = b'rec1'
kAENotifyStopRecording = b'rec0'
kAENotifyRecording = b'recr'

kAEUnknownSource = 0
kAEDirectCall = 1
kAESameProcess = 2
kAELocalProcess = 3
kAERemoteProcess = 4

# AEInteraction.h

kAEInteractWithSelf = 0
kAEInteractWithLocal = 1
kAEInteractWithAll = 2

kAEDoNotIgnoreHandler = 0x00000000
kAEIgnoreAppPhacHandler = 0x00000001
kAEIgnoreAppEventHandler = 0x00000002
kAEIgnoreSysPhacHandler = 0x00000004
kAEIgnoreSysEventHandler = 0x00000008
kAEIngoreBuiltInEventHandler = 0x00000010
kAEDontDisposeOnResume = 0x80000000

kAENoDispatch = 0
kAEUseStandardDispatch = 0xFFFFFFFF

# AppleScript.h

typeAppleScript = b'ascr'
kAppleScriptSubtype = typeAppleScript
typeASStorage = typeAppleScript

kASSelectInit = 0x1001
kASSelectSetSourceStyles = 0x1002
kASSelectGetSourceStyles = 0x1003
kASSelectGetSourceStyleNames = 0x1004
kASSelectCopySourceAttributes = 0x1005
kASSelectSetSourceAttributes = 0x1006

kASHasOpenHandler = b'hsod'

kASDefaultMinStackSize = 4
kASDefaultPreferredStackSize = 16
kASDefaultMaxStackSize = 16
kASDefaultMinHeapSize = 4
kASDefaultPreferredHeapSize = 16
kASDefaultMaxHeapSize = 32

kASSourceStyleUncompiledText = 0
kASSourceStyleNormalText = 1
kASSourceStyleLanguageKeyword = 2
kASSourceStyleApplicationKeyword = 3
kASSourceStyleComment = 4
kASSourceStyleLiteral = 5
kASSourceStyleUserSymbol = 6
kASSourceStyleObjectSpecifier = 7
kASNumberOfSourceStyles = 8

# ASDebugging.h

kOSAModeDontDefine = 0x0001

kASSelectSetPropertyObsolete = 0x1101
kASSelectGetPropertyObsolete = 0x1102
kASSelectSetHandlerObsolete = 0x1103
kASSelectGetHandlerObsolete = 0x1104
kASSelectGetAppTerminologyObsolete = 0x1105
kASSelectSetProperty = 0x1106
kASSelectGetProperty = 0x1107
kASSelectSetHandler = 0x1108
kASSelectGetHandler = 0x1109
kASSelectGetAppTerminology = 0x110A
kASSelectGetSysTerminology = 0x110B
kASSelectGetPropertyNames = 0x110C
kASSelectGetHandlerNames = 0x110D

# ASRegistry.h

keyAETarget = b'targ'
keySubjectAttr = b'subj'
keyASReturning = b'Krtn'
kASAppleScriptSuite = b'ascr'
kASScriptEditorSuite = b'ToyS'
kASTypeNamesSuite = b'tpnm'
typeAETE = b'aete'
typeAEUT = b'aeut'
kGetAETE = b'gdte'
kGetAEUT = b'gdut'
kUpdateAEUT = b'udut'
kUpdateAETE = b'udte'
kCleanUpAEUT = b'cdut'
kASComment = b'cmnt'
kASLaunchEvent = b'noop'
keyScszResource = b'scsz'
typeScszResource = b'scsz'
kASSubroutineEvent = b'psbr'
keyASSubroutineName = b'snam'
kASPrepositionalSubroutine = b'psbr'
keyASPositionalArgs = b'parg'

keyAppHandledCoercion = b'idas'

kASStartLogEvent = b'log1'
kASStopLogEvent = b'log0'
kASCommentEvent = b'cmnt'

kASAdd = b'+   '
kASSubtract = b'-   '
kASMultiply = b'*   '
kASDivide = b'/   '
kASQuotient = b'div '
kASRemainder = b'mod '
kASPower = b'^   '
kASEqual = kAEEquals
kASNotEqual = 0xAD202020
kASGreaterThan = kAEGreaterThan
kASGreaterThanOrEqual = kAEGreaterThanEquals
kASLessThan = kAELessThan
kASLessThanOrEqual = kAELessThanEquals
kASComesBefore = b'cbfr'
kASComesAfter = b'cafr'
kASConcatenate = b'ccat'
kASStartsWith = kAEBeginsWith
kASEndsWith = kAEEndsWith
kASContains = kAEContains

kASAnd = kAEAND
kASOr = kAEOR
kASNot = kAENOT
kASNegate = b'neg '
keyASArg = b'arg '

kASErrorEventCode = b'err '
kOSAErrorArgs = b'erra'
keyAEErrorObject = b'erob'
pLength = b'leng'
pReverse = b'rvse'
pRest = b'rest'
pInherits = b'c@#^'
pProperties = b'pALL'
keyASUserRecordFields = b'usrf'
typeUserRecordFields = typeAEList

keyASPrepositionAt = b'at  '
keyASPrepositionIn = b'in  '
keyASPrepositionFrom = b'from'
keyASPrepositionFor = b'for '
keyASPrepositionTo = b'to  '
keyASPrepositionThru = b'thru'
keyASPrepositionThrough = b'thgh'
keyASPrepositionBy = b'by  '
keyASPrepositionOn = b'on  '
keyASPrepositionInto = b'into'
keyASPrepositionOnto = b'onto'
keyASPrepositionBetween = b'btwn'
keyASPrepositionAgainst = b'agst'
keyASPrepositionOutOf = b'outo'
keyASPrepositionInsteadOf = b'isto'
keyASPrepositionAsideFrom = b'asdf'
keyASPrepositionAround = b'arnd'
keyASPrepositionBeside = b'bsid'
keyASPrepositionBeneath = b'bnth'
keyASPrepositionUnder = b'undr'

keyASPrepositionOver = b'over'
keyASPrepositionAbove = b'abve'
keyASPrepositionBelow = b'belw'
keyASPrepositionApartFrom = b'aprt'
keyASPrepositionGiven = b'givn'
keyASPrepositionWith = b'with'
keyASPrepositionWithout = b'wout'
keyASPrepositionAbout = b'abou'
keyASPrepositionSince = b'snce'
keyASPrepositionUntil = b'till'

kDialectBundleResType = b'Dbdl'
cConstant = typeEnumerated
cClassIdentifier = pClass
cObjectBeingExamined = typeObjectBeingExamined
cList = typeAEList
cSmallReal = typeIEEE32BitFloatingPoint
cReal = typeIEEE64BitFloatingPoint
cRecord = typeAERecord
cReference = cObjectSpecifier
cUndefined = b'undf'
cMissingValue = b'msng'
cSymbol = b'symb'
cLinkedList = b'llst'
cVector = b'vect'
cEventIdentifier = b'evnt'
cKeyIdentifier = b'kyid'
cUserIdentifier = b'uid '
cPreposition = b'prep'
cKeyForm = enumKeyForm
cScript = b'scpt'
cHandler = b'hand'
cProcedure = b'proc'

cHandleBreakpoint = b'brak'

cClosure = b'clsr'
cRawData = b'rdat'
cStringClass = typeChar
cNumber = b'nmbr'
cListElement = b'celm'
cListOrRecord = b'lr  '
cListOrString = b'ls  '
cListRecordOrString = b'lrs '
cNumberOrString = b'ns  '
cNumberOrDateTime = b'nd  '
cNumberDateTimeOrString = b'nds '
cAliasOrString = b'sf  '
cSeconds = b'scnd'
typeSound = b'snd '
enumBooleanValues = b'boov'
kAETrue = typeTrue
kAEFalse = typeFalse
enumMiscValues = b'misc'
kASCurrentApplication = b'cura'
formUserPropertyID = b'usrp'

cString = cStringClass

pASIt = b'it  '
pASMe = b'me  '
pASResult = b'rslt'
pASSpace = b'spac'
pASReturn = b'ret '
pASTab = b'tab '
pASPi = b'pi  '
pASParent = b'pare'
kASInitializeEventCode = b'init'
pASPrintLength = b'prln'
pASPrintDepth = b'prdp'
pASTopLevelScript = b'ascr'

kAECase = b'case'
kAEDiacritic = b'diac'
kAEWhiteSpace = b'whit'
kAEHyphens = b'hyph'
kAEExpansion = b'expa'
kAEPunctuation = b'punc'
kAEZenkakuHankaku = b'zkhk'
kAESmallKana = b'skna'
kAEKataHiragana = b'hika'
kASConsiderReplies = b'rmte'
kASNumericStrings = b'nume'
enumConsiderations = b'cons'

kAECaseConsiderMask = 0x00000001
kAEDiacriticConsiderMask = 0x00000002
kAEWhiteSpaceConsiderMask = 0x00000004
kAEHyphensConsiderMask = 0x00000008
kAEExpansionConsiderMask = 0x00000010
kAEPunctuationConsiderMask = 0x00000020
kASConsiderRepliesConsiderMask = 0x00000040
kASNumericStringsConsiderMask = 0x00000080
kAECaseIgnoreMask = 0x00010000
kAEDiacriticIgnoreMask = 0x00020000
kAEWhiteSpaceIgnoreMask = 0x00040000
kAEHyphensIgnoreMask = 0x00080000
kAEExpansionIgnoreMask = 0x00100000
kAEPunctuationIgnoreMask = 0x00200000
kASConsiderRepliesIgnoreMask = 0x00400000
kASNumericStringsIgnoreMask = 0x00800000
enumConsidsAndIgnores = b'csig'

cCoercion = b'coec'
cCoerceUpperCase = b'txup'
cCoerceLowerCase = b'txlo'
cCoerceRemoveDiacriticals = b'txdc'
cCoerceRemovePunctuation = b'txpc'
cCoerceRemoveHyphens = b'txhy'
cCoerceOneByteToTwoByte = b'txex'
# noinspection SpellCheckingInspection,SpellCheckingInspection
cCoerceRemoveWhiteSpace = b'txws'
cCoerceSmallKana = b'txsk'
cCoerceZenkakuhankaku = b'txze'
cCoerceKataHiragana = b'txkh'
cZone = b'zone'
cMachine = b'mach'
cAddress = b'addr'
cRunningAddress = b'radd'
cStorage = b'stor'

pASWeekday = b'wkdy'
pASMonth = b'mnth'
pASDay = b'day '
pASYear = b'year'
pASTime = b'time'
pASDateString = b'dstr'
pASTimeString = b'tstr'
cMonth = pASMonth
cJanuary = b'jan '
cFebruary = b'feb '
cMarch = b'mar '
cApril = b'apr '
cMay = b'may '
cJune = b'jun '
cJuly = b'jul '
cAugust = b'aug '
cSeptember = b'sep '
cOctober = b'oct '
cNovember = b'nov '
cDecember = b'dec '

cWeekday = pASWeekday
cSunday = b'sun '
cMonday = b'mon '
cTuesday = b'tue '
cWednesday = b'wed '
cThursday = b'thu '
cFriday = b'fri '
cSaturday = b'sat '
pASQuote = b'quot'
pASSeconds = b'secs'
pASMinutes = b'min '
pASHours = b'hour'
pASDays = b'days'
pASWeeks = b'week'
cWritingCodeInfo = b'citl'
pScriptCode = b'pscd'
pLangCode = b'plcd'
kASMagicTellEvent = b'tell'
kASMagicEndTellEvent = b'tend'

# DigitalHubRegistry.h

kDigiHubEventClass = b'dhub'

kDigiHubMusicCD = b'aucd'
kDigiHubPictureCD = b'picd'
kDigiHubVideoDVD = b'vdvd'
kDigiHubBlankCD = b'bcd '
kDigiHubBlankDVD = b'bdvd'

# OSA.h

kOSAComponentType = b'osa '

kOSAGenericScriptingComponentSubtype = b'scpt'

kOSAFileType = b'osas'

kOSASuite = b'ascr'

kOSARecordedText = b'recd'

kOSAScriptIsModified = b'modi'

kOSAScriptIsTypeCompiledScript = b'cscr'

kOSAScriptIsTypeScriptValue = b'valu'

kOSAScriptIsTypeScriptContext = b'cntx'

kOSAScriptBestType = b'best'

kOSACanGetSource = b'gsrc'

typeOSADialectInfo = b'difo'
keyOSADialectName = b'dnam'
keyOSADialectCode = b'dcod'
keyOSADialectLangCode = b'dlcd'
keyOSADialectScriptCode = b'dscd'

kOSANullScript = 0

kOSANullMode = 0
kOSAModeNull = 0

kOSASupportsCompiling = 0x0002
kOSASupportsGetSource = 0x0004
kOSASupportsAECoercion = 0x0008
kOSASupportsAESending = 0x0010
kOSASupportsRecording = 0x0020
kOSASupportsConvenience = 0x0040
kOSASupportsDialects = 0x0080
kOSASupportsEventHandling = 0x0100

kOSASelectLoad = 0x0001
kOSASelectStore = 0x0002
kOSASelectExecute = 0x0003
kOSASelectDisplay = 0x0004
kOSASelectScriptError = 0x0005
kOSASelectDispose = 0x0006
kOSASelectSetScriptInfo = 0x0007
kOSASelectGetScriptInfo = 0x0008
kOSASelectSetActiveProc = 0x0009
kOSASelectGetActiveProc = 0x000A
kOSASelectCopyDisplayString = 0x000B

kOSASelectScriptingComponentName = 0x0102
kOSASelectCompile = 0x0103
kOSASelectCopyID = 0x0104

kOSASelectCopyScript = 0x0105

kOSASelectGetSource = 0x0201
kOSASelectCopySourceString = 0x0202

kOSASelectCoerceFromDesc = 0x0301
kOSASelectCoerceToDesc = 0x0302

kOSASelectSetSendProc = 0x0401
kOSASelectGetSendProc = 0x0402
kOSASelectSetCreateProc = 0x0403
kOSASelectGetCreateProc = 0x0404
kOSASelectSetDefaultTarget = 0x0405

kOSASelectStartRecording = 0x0501
kOSASelectStopRecording = 0x0502

kOSASelectLoadExecute = 0x0601
kOSASelectCompileExecute = 0x0602
kOSASelectDoScript = 0x0603

kOSASelectSetCurrentDialect = 0x0701
kOSASelectGetCurrentDialect = 0x0702
kOSASelectAvailableDialects = 0x0703
kOSASelectGetDialectInfo = 0x0704
kOSASelectAvailableDialectCodeList = 0x0705

kOSASelectSetResumeDispatchProc = 0x0801
kOSASelectGetResumeDispatchProc = 0x0802
kOSASelectExecuteEvent = 0x0803
kOSASelectDoEvent = 0x0804
kOSASelectMakeContext = 0x0805

kOSASelectComponentSpecificStart = 0x1001

kOSAModePreventGetSource = 0x00000001

kOSAModeNeverInteract = kAENeverInteract
kOSAModeCanInteract = kAECanInteract
kOSAModeAlwaysInteract = kAEAlwaysInteract
kOSAModeDontReconnect = kAEDontReconnect

kOSAModeCantSwitchLayer = 0x00000040

kOSAModeDoRecord = 0x00001000

kOSAModeCompileIntoContext = 0x00000002

kOSAModeAugmentContext = 0x00000004

kOSAModeDisplayForHumans = 0x00000008

kOSAModeDontStoreParent = 0x00010000

kOSAModeDispatchToDirectObject = 0x00020000

kOSAModeDontGetDataForArguments = 0x00040000

kOSAModeFullyQualifyDescriptors = 0x00080000

kOSAScriptResourceType = kOSAGenericScriptingComponentSubtype

typeOSAGenericStorage = kOSAScriptResourceType

kOSAErrorNumber = keyErrorNumber

kOSAErrorMessage = keyErrorString

kOSAErrorBriefMessage = b'errb'

kOSAErrorApp = b'erap'

kOSAErrorPartialResult = b'ptlr'

kOSAErrorOffendingObject = b'erob'

kOSAErrorExpectedType = b'errt'

kOSAErrorRange = b'erng'

typeOSAErrorRange = b'erng'

keyOSASourceStart = b'srcs'

keyOSASourceEnd = b'srce'

kOSAUseStandardDispatch = kAEUseStandardDispatch

kOSANoDispatch = kAENoDispatch

kOSADontUsePhac = 0x0001

# OSAComp.h


# OSAGeneric.h

kGenericComponentVersion = 0x0100

kGSSSelectGetDefaultScriptingComponent = 0x1001
kGSSSelectSetDefaultScriptingComponent = 0x1002
kGSSSelectGetScriptingComponent = 0x1003
kGSSSelectGetScriptingComponentFromStored = 0x1004
kGSSSelectGenericToRealID = 0x1005
kGSSSelectRealToGenericID = 0x1006
kGSSSelectOutOfRange = 0x1007



from dataclasses import dataclass
from types import MappingProxyType
from typing import Dict, Tuple, Union

KB: int = 10 ** 3
MB: int = 10 ** 6
GB: int = 10 ** 9

SECOND: int = 1
MINUTE: int = 60 * SECOND
HOUR: int = 60 * MINUTE


class DNSMapping(str):
    pass


@dataclass(frozen=True)
class Point:
    x: Union[int, float]
    y: Union[int, float]


@dataclass(frozen=True)
class Frame:
    x1: Union[int, float]
    y1: Union[int, float]
    x2: Union[int, float]
    y2: Union[int, float]
    center: Point
    width: Union[int, float]
    height: Union[int, float]


@dataclass(frozen=True)
class Region:
    x1: Union[int, float]
    y1: Union[int, float]
    x2: Union[int, float]
    y2: Union[int, float]


@dataclass(frozen=True)
class ScreenSize:
    width: Union[int, float]
    height: Union[int, float]


@dataclass(frozen=True)
class CheckboxState:
    states: MappingProxyType = MappingProxyType({
        'unchecked': 0, 0: 0,
        'checked': 1, 1: 1,
        'indeterminate': 2, 2: 2}
    )
    checked: int = 1
    unchecked: int = 0
    indeterminate: int = 2


@dataclass(frozen=True)
class Colors:
    red: Tuple[str, ...] = ('lightcoral', 'lightsalmon', 'salmon', 'tomato',)
    green: Tuple[str, ...] = ('lightgreen', 'limegreen', 'mediumaquamarine', 'mediumseagreen', 'darkseagreen')
    blue: Tuple[str, ...] = ('deepskyblue', 'dodgerblue', 'cornflowerblue', 'royalblue',
                             'lavender', 'lightskyblue', 'mediumturquoise',)
    orange: Tuple[str, ...] = ('coral', 'lightsalmon', 'sandybrown', 'darkorange', 'goldenrod',)
    grey: Tuple[str, ...] = ('darkgray', 'darkslategray', 'dimgray', 'lightgray', 'slategray',
                             'darkgrey', 'darkslategrey', 'dimgrey', 'lightgrey', 'slategrey',
                             'silver',)


@dataclass(frozen=True)
class DisclosureTriangleState:
    closed: int = 0
    opened: int = 1


@dataclass(frozen=True)
class SettingsOpenCV:
    min_sim: float = 0.88
    wait_time: int = 5


@dataclass
class URLDetails:
    url: str
    schema: str
    domain: str
    path: str
    query: Dict[str, str]



from typing import List, Dict, Tuple, Optional


class SystemInformation:
    def __init__(self, executor):
        self.executor = executor
        self.__version: Optional[Tuple[int, ...]] = None
        self.__sphardwaredatatype: Optional[Dict[str, str]] = None

    @property
    def uptime(self) -> int:
        timestamp = self.executor.get_output('date +%s')
        boot_time = self.executor.get_output("sysctl -n kern.boottime | awk '{print $4}' | sed 's/,//g'")
        return int(timestamp) - int(boot_time)

    @property
    def iousb_devices(self) -> List[str]:
        command: str = "ioreg -p IOUSB -w0 | sed 's/[^o]*o //; s/@.*$//' |grep -v '^Root.*'"
        return self.executor.get_output(command).splitlines()

    @property
    def serial(self) -> str:
        return self._sphardwaredatatype.get('Serial Number')

    @property
    def computer_name(self) -> str:
        return self.executor.get_output('scutil --get ComputerName')

    @property
    def model_name(self) -> str:
        return self._sphardwaredatatype.get('Model Name')

    @property
    def hw_uuid(self) -> str:
        return self._sphardwaredatatype.get('Hardware UUID')

    @property
    def mac_model(self) -> str:
        return self._sphardwaredatatype.get('Model Identifier')

    @property
    def uuid(self) -> str:
        return self.executor.get_output('dsmemberutil getuuid -U ${USER}')

    @property
    def _sphardwaredatatype(self) -> Dict[str, str]:
        """'Serial Number', 'Model Identifier', 'Boot ROM Version', 'L3 Cache', 'Processor Name',
        'Total Number of Cores', 'Number of Processors', 'Hardware UUID', 'SMC Version', 'Model Name',
        'Memory', 'L2 Cache (per Core)', 'Processor Speed'."""
        if self.__sphardwaredatatype is None:
            output = self.executor.get_output('system_profiler SPHardwareDataType')
            items = (line.strip() for line in output[output.find('Model Name'):].splitlines())
            _sphardwaredatatype = dict(self.__process_sphardwaredatatype_item(item) for item in items)
            self.__sphardwaredatatype = _sphardwaredatatype
        return self.__sphardwaredatatype

    @staticmethod
    def __process_sphardwaredatatype_item(item) -> Tuple[str, str]:
        return tuple(item.replace(' (system)', '').split(': ')[:2])



import time

import AppKit
import Quartz
import pytweening


class MouseController:
    def __init__(self, tween_type=pytweening.easeInOutQuad):
        self.tween_type = tween_type
        self.__screen_size = None

    def move_to(self, x: int, y: int, duration: float = .25):
        self.__mouse_move_drag(x=x, y=y, duration=duration)
        time.sleep(.1)

    def drag_to(self, x: int, y: int, duration: float = .25):
        self.__mouse_move_drag(x=x, y=y, duration=duration, move='drag')
        time.sleep(.1)

    def mouse_down(self, x: int, y: int, button: str):
        if button == 'left':
            self._send_mouse_event(Quartz.kCGEventLeftMouseDown, x, y, Quartz.kCGMouseButtonLeft)
        elif button == 'middle':
            self._send_mouse_event(Quartz.kCGEventOtherMouseDown, x, y, Quartz.kCGMouseButtonCenter)
        elif button == 'right':
            self._send_mouse_event(Quartz.kCGEventRightMouseDown, x, y, Quartz.kCGMouseButtonRight)
        else:
            raise ValueError("button argument not in ('left', 'middle', 'right')")

    def mouse_up(self, x: int, y: int, button: str):
        if button == 'left':
            self._send_mouse_event(Quartz.kCGEventLeftMouseUp, x, y, Quartz.kCGMouseButtonLeft)
        elif button == 'middle':
            self._send_mouse_event(Quartz.kCGEventOtherMouseUp, x, y, Quartz.kCGMouseButtonCenter)
        elif button == 'right':
            self._send_mouse_event(Quartz.kCGEventRightMouseUp, x, y, Quartz.kCGMouseButtonRight)
        else:
            raise ValueError("button argument not in ('left', 'middle', 'right')")

    def __mouse_move_drag(self, x: int, y: int, duration: float, move: str = 'move'):
        kcg_event, mouse_button = Quartz.kCGEventMouseMoved, 0
        if move == 'drag':
            kcg_event, mouse_button = Quartz.kCGEventLeftMouseDragged, Quartz.kCGMouseButtonLeft
        start_x, start_y = self._position
        width, height = self.screen_size
        x = max(0, min(x, width - 1))  # Make sure x and y are within the screen bounds.
        y = max(0, min(y, height - 1))
        steps_count = int(max(abs(x - start_x), abs(y - start_y))) or 1
        pause = duration / steps_count
        steps = [pytweening.getPointOnLine(start_x, start_y, x, y, self.tween_type(n / steps_count))
                 for n in range(steps_count)] + [(x, y)]
        if steps_count < 20:
            steps, pause = [(x, y)], 0
        for tween_x, tween_y in steps:
            self._send_mouse_event(kcg_event, tween_x, tween_y, mouse_button)
            time.sleep(pause)

    @staticmethod
    def vertical_scroll(scrolls: int):
        for _ in range(scrolls):
            swe = Quartz.CGEventCreateScrollWheelEvent(None, Quartz.kCGScrollEventUnitLine, 1, 10)
            Quartz.CGEventPost(Quartz.kCGHIDEventTap, swe)
            time.sleep(0.03)

    @staticmethod
    def multi_click(x: int, y: int, button: str, clicks: int):
        if button == 'left':
            btn = Quartz.kCGMouseButtonLeft
            down = Quartz.kCGEventLeftMouseDown
            up = Quartz.kCGEventLeftMouseUp
        elif button == 'middle':
            btn = Quartz.kCGMouseButtonCenter
            down = Quartz.kCGEventOtherMouseDown
            up = Quartz.kCGEventOtherMouseUp
        elif button == 'right':
            btn = Quartz.kCGMouseButtonRight
            down = Quartz.kCGEventRightMouseDown
            up = Quartz.kCGEventRightMouseUp
        else:
            raise ValueError("button argument not in ('left', 'middle', 'right')")

        mouse_event = Quartz.CGEventCreateMouseEvent(None, down, (x, y), btn)
        Quartz.CGEventSetIntegerValueField(mouse_event, Quartz.kCGMouseEventClickState, clicks)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, mouse_event)
        Quartz.CGEventSetType(mouse_event, up)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, mouse_event)
        for _ in range(clicks - 1):
            Quartz.CGEventSetType(mouse_event, down)
            Quartz.CGEventPost(Quartz.kCGHIDEventTap, mouse_event)
            Quartz.CGEventSetType(mouse_event, up)
            Quartz.CGEventPost(Quartz.kCGHIDEventTap, mouse_event)

    @staticmethod
    def _send_mouse_event(event, x: int, y: int, button):
        event = Quartz.CGEventCreateMouseEvent(None, event, (x, y), button)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, event)

    @property
    def _position(self):
        loc = AppKit.NSEvent.mouseLocation()
        return int(loc.x), int(Quartz.CGDisplayPixelsHigh(0) - loc.y)

    @property
    def screen_size(self):
        if self.__screen_size is None:
            self.__screen_size = Quartz.CGDisplayPixelsWide(Quartz.CGMainDisplayID()), Quartz.CGDisplayPixelsHigh(
                Quartz.CGMainDisplayID())
        return self.__screen_size



import time

import AppKit
import Quartz

from macuitest.lib.elements.controllers.keyboard_mappings import KEYBOARD_KEYS, SPECIAL_KEYS


class KeyBoardController:
    def __init__(self):
        pass

    def write(self, message: str, pause: float = .001):
        """Press key for each of the characters in message."""
        for char in message:
            self.__send_key_event(char, 'down')
            time.sleep(pause)
            self.__send_key_event(char, 'up')
            time.sleep(pause)

    def hotkey(self, *args):
        """Calling `hotkey('command', 'shift', 'a')` performs a "CMD-Shift-A" shortcut press."""
        for c in args:
            if len(c) > 1:
                c = c.lower()
            self.__send_key_event(c, 'down')
            time.sleep(.025)
        for c in reversed(args):
            if len(c) > 1:
                c = c.lower()
            self.__send_key_event(c, 'up')
            time.sleep(.025)

    def __send_key_event(self, key: str, event: str):
        send = self.send_special_key_event if key in SPECIAL_KEYS else self.send_regular_key_event
        send(key, event)

    def send_regular_key_event(self, key: str, event_type):
        if KEYBOARD_KEYS.get(key) is None:
            raise ValueError(f'Key "{key}" is not available')
        if self.is_shift_char(key):
            key_code = KEYBOARD_KEYS[key.lower()]
            event = Quartz.CGEventCreateKeyboardEvent(None, KEYBOARD_KEYS['shift'], event_type == 'down')
            Quartz.CGEventPost(Quartz.kCGHIDEventTap, event)
            time.sleep(.03)  # Tiny sleep to let OS X catch up on us pressing shift
        else:
            key_code = KEYBOARD_KEYS[key]
        event = Quartz.CGEventCreateKeyboardEvent(None, key_code, event_type == 'down')
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, event)

    @staticmethod
    def send_special_key_event(key, event_type):
        """Helper method for special keys."""
        key_code = SPECIAL_KEYS[key]
        ev = AppKit.NSEvent.otherEventWithType_location_modifierFlags_timestamp_windowNumber_context_subtype_data1_data2_(
            Quartz.NSSystemDefined,  # type
            (0, 0),  # location
            0xa00 if event_type == 'down' else 0xb00,  # flags
            0,  # timestamp
            0,  # window
            0,  # ctx
            8,  # subtype
            (key_code << 16) | ((0xa if event_type == 'down' else 0xb) << 8),  # data1
            -1  # data2
        )
        Quartz.CGEventPost(0, ev.CGEvent())

    @staticmethod
    def is_shift_char(character: str):
        """Returns True if the key character is uppercase or shifted."""
        return character.isupper() or character in '~!@#$%^&*()_+{}|:"<>?'


keyboard = KeyBoardController()



import threading
from typing import List, Dict, Optional

import biplist

from macuitest.lib import core


class MemoryManager:
    """Interface to memory (RAM) information and manipulations on it."""

    def __init__(self, executor):
        self._memory_slots: Optional[List[Dict[str, str]]] = None
        self.executor = executor

    def load_memory(self, percent_free: int, timeout: int = 15) -> None:
        """Simulate memory pressure on the system in background."""
        threading.Thread(target=self.executor.execute, args=(f'memory_pressure -p {percent_free}', timeout)).start()

    @property
    def total_memory_bytes(self) -> int:
        return core.convert_file_size_to_bytes(self.total_memory_str)

    @property
    def total_memory_str(self) -> str:
        return f'{self.total_memory} GB'

    @property
    def total_memory(self) -> int:
        return sum((int(bank.get('dimm_size').split()[0]) for bank in self.memory_slots))

    @property
    def memory_banks(self) -> int:
        return len(self.memory_slots)

    @property
    def memory_speed(self) -> str:
        return self.memory_slots[0].get('dimm_speed')

    @property
    def memory_type(self) -> str:
        return self.memory_slots[0].get('dimm_type')

    @property
    def memory_slots(self) -> List[Dict[str, str]]:
        if self._memory_slots is not None:
            return self._memory_slots
        _info = bytes(self.executor.get_output('system_profiler -xml SPMemoryDataType'), encoding='utf-8')
        _slots = biplist.readPlistFromString(_info)[0].get('_items')[0].get('_items')
        self._memory_slots = _slots
        return _slots



import re
from collections import namedtuple

from ApplicationServices import (AXUIElementGetTypeID, AXValueGetType, NSPointFromString, NSRangeFromString,
                                 NSSizeFromString, kAXValueCFRangeType, kAXValueCGPointType, kAXValueCGSizeType, )
from CoreFoundation import CFArrayGetTypeID, CFGetTypeID, CFStringGetTypeID


class Converter:
    def __init__(self, ax_ui_element_class=None):
        self.app_ref_class = ax_ui_element_class

    def convert_value(self, value):
        if CFGetTypeID(value) == CFStringGetTypeID():
            try:
                return str(value)
            except UnicodeEncodeError:
                return str(value.encode('utf-8'))
        if CFGetTypeID(value) == AXUIElementGetTypeID():
            return self.convert_app_ref(value)
        if CFGetTypeID(value) == CFArrayGetTypeID():
            return self.convert_list(value)
        if AXValueGetType(value) == kAXValueCGSizeType:
            return self.convert_size(value)
        if AXValueGetType(value) == kAXValueCGPointType:
            return self.convert_point(value)
        if AXValueGetType(value) == kAXValueCFRangeType:
            return self.convert_range(value)
        else:
            return value

    def convert_list(self, value):
        return [self.convert_value(item) for item in value]

    def convert_app_ref(self, value):
        return self.app_ref_class(ref=value)

    @staticmethod
    def convert_size(value):
        repr_searched = re.search('{.*}', str(value)).group()
        CGSize = namedtuple('CGSize', ['width', 'height'])
        size = NSSizeFromString(repr_searched)
        return CGSize(size.width, size.height)

    @staticmethod
    def convert_point(value):
        repr_searched = re.search('{.*}', str(value)).group()
        CGPoint = namedtuple('CGPoint', ['x', 'y'])
        point = NSPointFromString(repr_searched)
        return CGPoint(point.x, point.y)

    @staticmethod
    def convert_range(value):
        repr_searched = re.search('{.*}', str(value)).group()
        CFRange = namedtuple('CFRange', ['location', 'length'])
        _range_ = NSRangeFromString(repr_searched)
        return CFRange(_range_.location, _range_.length)



import time
from typing import Union, Tuple, ClassVar

from macuitest.lib.operating_system.env import env


class Defaults:
    preferences_daemon: ClassVar[str] = 'cfprefsd'

    def __init__(self, executor, service_manager):
        self.executor = executor
        self.service_manager = service_manager

    def clear_application_preferences(self, bundle_ids: Tuple[str, ...]):
        """Clear application preferences.
            :param bundle_ids: Tuple of target bundle ids."""
        for bundle_id in bundle_ids:
            self.remove_property_list_by_id(bundle_id, flush_pref_cache=False)
            self.executor.execute(f'rm -f {env.user_preferences}/{bundle_id}*')

    def remove_property_list_by_id(self, bundle_id, flush_pref_cache=True):
        self.executor.execute(f'defaults delete {bundle_id}')
        if flush_pref_cache:
            self.flush_preferences_cache()

    def read(self, domain: str, property_: str = 'MKIntroWasShown') -> str:
        return self.executor.get_output(f'defaults read {domain} {property_}')

    def delete_defaults(self, domain: str, property_: str = 'MKIntroWasShown'):
        self.executor.execute(f'defaults delete {domain} {property_}')
        time.sleep(2)

    def set_defaults(self, domain: str, property_: str = 'MKIntroWasShown 1', pause: int = 2):
        self.executor.execute(f'defaults write {domain} {property_}')
        time.sleep(pause)

    def flush_preferences_cache(self, pause: Union[int, float] = 3):
        time.sleep(pause)
        self.service_manager.kill_process(self.preferences_daemon)
        time.sleep(pause)



import time
from typing import Any, Optional

from macuitest.config.constants import Frame, Point, Region
from macuitest.lib import core
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.native.errors import AXErrorFactory, AXErrorInvalidUIElement
from macuitest.lib.elements.controllers.mouse import mouse


class NativeElement:
    def __init__(self, item):
        self.item = item

    def _select(self):
        self.item.AXSelected = True

    def _press(self, pause: float = 0.375):
        time.sleep(pause)
        self.item.AXPress()
        time.sleep(.2)

    def double_click_mouse(self, x_off=0, y_off=0, duration=.2):
        mouse.double_click(self.frame.center.x + x_off, self.frame.center.y + y_off, duration)

    def click_mouse(self, x_off=0, y_off=0, duration=.2, hold_time=.3):
        mouse.click(self.frame.center.x + x_off, self.frame.center.y + y_off, hold_time, duration)

    def rightclick_mouse(self, x_off=0, y_off=0, duration=.2):
        mouse.right_click(self.frame.center.x + x_off, self.frame.center.y + y_off, duration)

    def hover_mouse(self, x_off=0, y_off=0, duration=.2):
        mouse.hover(self.frame.center.x + x_off, self.frame.center.y + y_off, duration)

    def region(self, margin: int = 4):
        return Region(self.frame.x1 - margin, self.frame.y1 - margin, self.frame.x2 + margin, self.frame.y2 + margin)

    @property
    def frame(self) -> Frame:
        _frame = [*self.item.AXPosition, *self.item.AXSize]
        x1, y1, width, height = _frame
        x2, y2 = x1 + width, y1 + height
        center = Point(int((x1 + width / 2)), int((y1 + height / 2)))
        return Frame(x1, y1, x2, y2, center, width, height)

    @property
    def characters_number(self) -> int:
        return self.item.AXNumberOfCharacters

    @property
    def title(self) -> str:
        return self.item.AXTitle

    @property
    def help(self) -> str:
        return self.item.AXHelp

    @property
    def parent(self):
        return NativeElement(self.item.AXParent)

    @property
    def children(self):
        try:
            return self.item.AXChildren
        except AttributeError:
            return []

    @property
    def value(self) -> Any:
        return self.item.AXValue

    @value.setter
    def value(self, value_) -> None:
        self.value = value_

    @property
    def is_visible(self) -> bool:
        return self.exists

    def wait_vanish(self, timeout: [int, float] = 5) -> bool:
        return wait_condition(lambda: self.__get_axrole() is None, timeout=timeout)

    @property
    def did_vanish(self) -> bool:
        return wait_condition(lambda: self.__get_axrole() is None)

    def __get_axrole(self) -> Optional[str]:
        try:
            return self.item.AXRole
        except (AttributeError, IndexError):
            pass

    @property
    def exists(self) -> bool:
        return self.item is not None


class Clickable(NativeElement):
    @property
    def is_enabled(self) -> bool:
        return self.item.AXEnabled

    def press(self, pause: float = .375):
        self._press(pause=pause)

    def click(self, pause: float = .375):
        self._press(pause=pause)


class Cell(NativeElement):
    def select(self):
        self._select()


class Row(NativeElement):
    def select(self):
        self._select()

    @property
    def is_selected(self) -> bool:
        return self.item.AXSelected

    @is_selected.setter
    def is_selected(self, value):
        self.item.AXSelected = value


class Button(Clickable):
    pass


class Image(NativeElement):
    @property
    def label(self) -> str:
        return self.item.AXLabel


class StaticText(NativeElement):
    @property
    def to_bytes(self) -> int:
        return core.convert_file_size_to_bytes(self.text)

    @property
    def text(self) -> str:
        return str(self.value)

    def __eq__(self, other):
        return wait_condition(lambda: self.text == other, timeout=2)


class TextField(StaticText):
    @property
    def text(self) -> str:
        return str(self.item.AXValue)

    @text.setter
    def text(self, value):
        self.item.AXValue = value

    @property
    def keyboard_focused(self) -> bool:
        return self.item.AXFocused

    @keyboard_focused.setter
    def keyboard_focused(self, value):
        self.item.AXFocused = bool(value)

    def __eq__(self, other):
        return wait_condition(lambda: self.text == other, timeout=2)


class CheckBox(Clickable):
    @property
    def state(self) -> int:
        return self.value

    @state.setter
    def state(self, value: bool) -> None:
        for _ in range(3):
            if self.state != value:
                self._press()


class DisclosureTriangle(CheckBox):
    pass


class MenuItem(Clickable):
    pass


class Link(NativeElement):

    @property
    def url(self) -> str:
        return str(self.item.AXURL)


class WebView(NativeElement):

    @property
    def url(self) -> str:
        webview = self.__perform_lookup()
        wait_condition(lambda: webview.AXURL, timeout=30)
        wait_condition(lambda: webview.AXURL.startswith('https://'), timeout=30)
        return str(webview.AXURL)

    def __perform_lookup(self):
        self.__wait_children()
        return self.item.find_element(AXRole='AXUnknown', recursive=True)

    def __wait_children(self):
        for _ in range(20):
            try:
                time.sleep(0.3)
                return self.item.AXChildren
            except (AttributeError, AXErrorInvalidUIElement, AXErrorFactory):
                continue



from typing import Dict, ClassVar

from macuitest.lib.operating_system.db_manager import DataBaseManager
from macuitest.lib.operating_system.env import env


class TCCManager:
    """
    TCCManager manages the privacy database, which stores decisions
    the user has made about whether apps may access personal data.
    """
    sys_db: ClassVar[str] = f'{env.system_app_support}/com.apple.TCC/TCC.db'
    usr_db: ClassVar[str] = f'{env.home}{sys_db}'
    services: ClassVar[Dict[str, str]] = {
        'full_disk_access': 'kTCCServiceSystemPolicyAllFiles',
        'camera_access': 'kTCCServiceCamera',
        'documents_folder_access': 'kTCCServiceSystemPolicyDocumentsFolder',
        'desktop_folder_access': 'kTCCServiceSystemPolicyDesktopFolder',
        'downloads_folder_access': 'kTCCServiceSystemPolicyDownloadsFolder',
    }

    def __init__(self, domain: str = 'system'):
        self.db: str = self.sys_db if domain == 'system' else self.usr_db
        self.db_manager = DataBaseManager(self.db)

    def list_access_policies(self, client: str) -> Dict[str, int]:
        query = f'SELECT service, allowed FROM access WHERE client is "{client}"'
        return dict(self.db_manager.cursor.execute(query).fetchall())

    def revoke_access_to_service(self, service: str, client: str) -> None:
        query = f'DELETE FROM access WHERE service is "{service}" AND client is "{client}"'
        self.db_manager.cursor.execute(query)
        self.db_manager.connection.commit()



import time
from dataclasses import dataclass

from macuitest.lib.elements.controllers.keyboard_controller import KeyBoardController
from macuitest.lib.elements.controllers.mouse_controller import MouseController


@dataclass(frozen=True)
class Options:
    """Mouse options."""
    move: float = .25  # Cursor roaming time.
    hold: float = .25  # Time to hold a button selected.
    pause: float = .25  # Pause after an action.
    default_position: tuple = (28, 30)


class Mouse:
    """Cursor manipulator."""

    def __init__(self, controller: MouseController):
        self.controller = controller

    def double_click(self, x, y, _x=0, _y=0, duration=Options.move):
        """Hover over position and click twice."""
        self.click(x, y, _x, _y, duration=duration)
        self.controller.multi_click(x + _x, y + _y, button='left', clicks=2)

    def drag(self, x1: int, y1: int, x2: int, y2: int):
        self.hover(x1, y1)
        self.controller.mouse_down(x1, y1, 'left')
        self.controller.drag_to(x2, y2)
        self.controller.mouse_up(x2, y2, 'left')

    def click(self, x, y, _x=0, _y=0, hold=Options.hold, duration=Options.move, pause=Options.pause):
        """Hover over position and left-click once."""
        self.hover(x, y, _x, _y, duration)
        self._press_mouse_button(x + _x, y + _y, mouse_button='left', hold=hold, pause=pause)

    def right_click(self, x, y, _x=0, _y=0, hold=Options.hold, duration=Options.move, pause=Options.pause):
        """Hover over the position and control-click once."""
        self.hover(x, y, _x, _y, duration)
        self._press_mouse_button(x + _x, y + _y, mouse_button='right', hold=hold, pause=pause)

    def paste(self, x, y, _x=0, _y=0, phrase=''):
        """Hover over the position and click once. Then paste `phrase` from clipboard."""
        self.double_click(x + _x, y + _y)
        time.sleep(.5)
        KeyBoardController().write(phrase, pause=.005)

    def scroll(self, x, y, scrolls, _x=0, _y=0):
        self.hover(x, y, _x, _x)
        self.controller.vertical_scroll(scrolls)

    def reset(self):
        self.hover(*Options.default_position, duration=0.03)

    def hover(self, x: int, y: int, _x: int = 0, _y: int = 0, duration: float = Options.move) -> None:
        """Hover over the position."""
        self.controller.move_to(x + _x, y + _y, duration=duration)

    def _press_mouse_button(self, x: int, y: int, mouse_button: str, hold: float, pause: float):
        self.controller.mouse_down(x, y, mouse_button)
        time.sleep(hold)
        self.controller.mouse_up(x, y, mouse_button)
        time.sleep(pause)


mouse = Mouse(MouseController())



import time
from typing import Any, Callable, Union, Tuple


def wait_condition(predicate: Callable, timeout: Union[int, float] = 10, *args, **kwargs) -> Any:
    """Call `predicate` and return its result."""
    end = time.time() + timeout
    while time.time() < end:
        result = predicate(*args, **kwargs)
        if result:
            return result
        time.sleep(0.05)
    return False


def convert_version_from_tuple_to_str(string: Tuple[int, ...]) -> str:
    return '.'.join(str(i) for i in string)


def convert_version_from_str_to_tuple(string: str) -> Tuple[int, ...]:
    return tuple([int(item) for item in string.split('.')])


def convert_file_size_to_bytes(string: str) -> int:
    _size, _unit = string.strip().lower().replace('zero', '0').replace(',', '').split()
    dots = len([i for i in _size if i == '.'])
    return int(float(_size.replace('.', '', dots - 1)) * 1000 ** ('bytes', 'kb', 'mb', 'gb', 'tb').index(_unit))



import os
from pathlib import Path
from typing import Union

from macuitest.lib.operating_system.env import env


class ScreenshotPathBuilder:
    """Screenshot path builder."""

    def __init__(self, category: str, root: Union[Path, str] = os.environ.get('MACUITEST_SCR', os.environ.get('HOME'))):
        self.root = Path(root)
        self.category = category.lower().replace(' ', '_').replace('-', '_')

    def __getattr__(self, item: str):
        return getattr(self, item) if item == 'category' else self.build_path(self.category, item)

    def build_path(self, section: str, scr_name: str) -> str:
        """Build absolute path to a screenshot."""
        _base = self.root.joinpath(section).joinpath(f'{scr_name}.png')
        _macos_specific = self.root.joinpath(section).joinpath(f'{scr_name}_{env.version[1]}.png')
        _path = _macos_specific if _macos_specific.exists() else _base
        return _path.as_posix()



"""List of ISO 3166 alpha-2 country codes."""

country_codes = {
    'Afghanistan': 'AF',
    'Albania': 'AL',
    'Algeria': 'DZ',
    'American Samoa': 'AS',
    'Andorra': 'AD',
    'Angola': 'AO',
    'Anguilla': 'AI',
    'Antarctica': 'AQ',
    'Antigua and Barbuda': 'AG',
    'Argentina': 'AR',
    'Armenia': 'AM',
    'Aruba': 'AW',
    'Australia': 'AU',
    'Austria': 'AT',
    'Azerbaijan': 'AZ',
    'Bahamas': 'BS',
    'Bahrain': 'BH',
    'Bangladesh': 'BD',
    'Barbados': 'BB',
    'Belarus': 'BY',
    'Belgium': 'BE',
    'Belize': 'BZ',
    'Benin': 'BJ',
    'Bermuda': 'BM',
    'Bhutan': 'BT',
    'Bosnia and Herzegovina': 'BA',
    'Botswana': 'BW',
    'Brazil': 'BR',
    'British Indian Ocean Territory': 'IO',
    'Bulgaria': 'BG',
    'Burundi': 'BI',
    'Cambodia': 'KH',
    'Cameroon': 'CM',
    'Canada': 'CA',
    'Cape Verde': 'CV',
    'Cayman Islands': 'KY',
    'Central African Republic': 'CF',
    'Chad': 'TD',
    'Chile': 'CL',
    'China': 'CN',
    'Christmas Island': 'CX',
    'Cocos (Keeling) Islands': 'CC',
    'Colombia': 'CO',
    'Comoros': 'KM',
    'Congo': 'CD',
    'Cook Islands': 'CK',
    'Costa Rica': 'CR',
    'Croatia': 'HR',
    'Cuba': 'CU',
    'Cyprus': 'CY',
    'Czech Republic': 'CZ',
    'Denmark': 'DK',
    'Djibouti': 'DJ',
    'Dominica': 'DM',
    'Dominican Republic': 'DO',
    'Ecuador': 'EC',
    'Egypt': 'EG',
    'El Salvador': 'SV',
    'Equatorial Guinea': 'GQ',
    'Eritrea': 'ER',
    'Estonia': 'EE',
    'Ethiopia': 'ET',
    'Falkland Islands': 'FK',
    'Fiji': 'FJ',
    'Finland': 'FI',
    'France': 'FR',
    'French Guiana': 'GF',
    'French Polynesia': 'PF',
    'French Southern Territories': 'TF',
    'Gabon': 'GA',
    'Gambia': 'GM',
    'Georgia': 'GE',
    'Germany': 'DE',
    'Ghana': 'GH',
    'Gibraltar': 'GI',
    'Greece': 'GR',
    'Greenland': 'GL',
    'Grenada': 'GD',
    'Guadeloupe': 'GP',
    'Guam': 'GU',
    'Guatemala': 'GT',
    'Guernsey': 'GG',
    'Guinea': 'GN',
    'Guinea-Bissau': 'GW',
    'Guyana': 'GY',
    'Haiti': 'HT',
    'Heard Island and McDonald Islands': 'HM',
    'Holy See (Vatican City State)': 'VA',
    'Honduras': 'HN',
    'Hong Kong': 'HK',
    'Hungary': 'HU',
    'Iceland': 'IS',
    'India': 'IN',
    'Indonesia': 'ID',
    'Iran': 'IR',
    'Iraq': 'IQ',
    'Ireland': 'IE',
    'Isle of Man': 'IM',
    'Israel': 'IL',
    'Italy': 'IT',
    'Jamaica': 'JM',
    'Japan': 'JP',
    'Jersey': 'JE',
    'Jordan': 'JO',
    'Kazakhstan': 'KZ',
    'Kenya': 'KE',
    'Kiribati': 'KI',
    'South Korea': 'KR',
    'Kuwait': 'KW',
    'Kyrgyzstan': 'KG',
    'Latvia': 'LV',
    'Lebanon': 'LB',
    'Lesotho': 'LS',
    'Liberia': 'LR',
    'Libya': 'LY',
    'Liechtenstein': 'LI',
    'Lithuania': 'LT',
    'Luxembourg': 'LU',
    'Macao': 'MO',
    'Macedonia': 'MK',
    'Madagascar': 'MG',
    'Malawi': 'MW',
    'Malaysia': 'MY',
    'Maldives': 'MV',
    'Mali': 'ML',
    'Malta': 'MT',
    'Marshall Islands': 'MH',
    'Martinique': 'MQ',
    'Mauritania': 'MR',
    'Mauritius': 'MU',
    'Mexico': 'MX',
    'Micronesia': 'FM',
    'Moldova': 'MD',
    'Monaco': 'MC',
    'Mongolia': 'MN',
    'Montenegro': 'ME',
    'Montserrat': 'MS',
    'Morocco': 'MA',
    'Mozambique': 'MZ',
    'Myanmar': 'MM',
    'Namibia': 'NA',
    'Nauru': 'NR',
    'Nepal': 'NP',
    'New Caledonia': 'NC',
    'New Zealand': 'NZ',
    'Nicaragua': 'NI',
    'Niger': 'NE',
    'Nigeria': 'NG',
    'Norfolk Island': 'NF',
    'Northern Mariana Islands': 'MP',
    'Norway': 'NO',
    'Oman': 'OM',
    'Pakistan': 'PK',
    'Panama': 'PA',
    'Paraguay': 'PY',
    'Peru': 'PE',
    'Philippines': 'PH',
    'Pitcairn': 'PN',
    'Poland': 'PL',
    'Portugal': 'PT',
    'Puerto Rico': 'PR',
    'Qatar': 'QA',
    'Romania': 'RO',
    'Russia': 'RU',
    'Russian Federation': 'RU',
    'Rwanda': 'RW',
    'Saint Lucia': 'LC',
    'Saint Martin (French part)': 'MF',
    'Saint Vincent and the Grenadines': 'VC',
    'Samoa': 'WS',
    'San Marino': 'SM',
    'Sao Tome and Principe': 'ST',
    'Saudi Arabia': 'SA',
    'Senegal': 'SN',
    'Serbia': 'RS',
    'Seychelles': 'SC',
    'Sierra Leone': 'SL',
    'Singapore': 'SG',
    'Slovakia': 'SK',
    'Slovenia': 'SI',
    'Solomon Islands': 'SB',
    'Somalia': 'SO',
    'South Africa': 'ZA',
    'South Georgia and the South Sandwich Islands': 'GS',
    'South Sudan': 'SS',
    'Spain': 'ES',
    'Sri Lanka': 'LK',
    'Sudan': 'SD',
    'Suriname': 'SR',
    'Swaziland': 'SZ',
    'Sweden': 'SE',
    'Switzerland': 'CH',
    'Syrian Arab Republic': 'SY',
    'Taiwan': 'TW',
    'Tajikistan': 'TJ',
    'Tanzania': 'TZ',
    'Thailand': 'TH',
    'The Netherlands': 'NL',
    'Togo': 'TG',
    'Tonga': 'TO',
    'Trinidad and Tobago': 'TT',
    'Tunisia': 'TN',
    'Turkey': 'TR',
    'Turkmenistan': 'TM',
    'Tuvalu': 'TV',
    'Uganda': 'UG',
    'Ukraine': 'UA',
    'United Arab Emirates': 'AE',
    'United Kingdom': 'GB',
    'United States': 'US',
    'United States Minor Outlying Islands': 'UM',
    'Uruguay': 'UY',
    'Uzbekistan': 'UZ',
    'Vanuatu': 'VU',
    'Venezuela': 'VE',
    'Vietnam': 'VN',
    'Virgin Islands': 'VI',
    'Western Sahara': 'EH',
    'Yemen': 'YE',
    'Zambia': 'ZM',
    'Zimbabwe': 'ZW',
}



import fnmatch
from typing import Optional


class LookUpMixin:
    def find_element(self, **kwargs):
        """Return the first object that matches the criteria."""
        return self._find_element(**kwargs)

    def find_elements(self, **kwargs):
        """Return a list of all children that match the specified criteria."""
        return list(self._find_elements(**kwargs))

    def text_areas(self, title: Optional[str] = None):
        return self._convenience_match('AXTextArea', 'AXTitle', title)

    def text_areas_r(self, title: Optional[str] = None):
        return self._convenience_match('AXTextArea', 'AXTitle', title, recursive=True)

    def text_fields(self, role_description: Optional[str] = None):
        return self._convenience_match('AXTextField', 'AXRoleDescription', role_description)

    def text_fields_r(self, role_description: Optional[str] = None):
        return self._convenience_match('AXTextField', 'AXRoleDescription', role_description, recursive=True)

    def buttons(self, title: Optional[str] = None):
        return self._convenience_match('AXButton', 'AXTitle', title)

    def buttons_r(self, title: Optional[str] = None):
        return self._convenience_match('AXButton', 'AXTitle', title, recursive=True)

    def windows(self, title: Optional[str] = None):
        return self._convenience_match('AXWindow', 'AXTitle', title)

    def windows_r(self, title: Optional[str] = None):
        return self._convenience_match('AXWindow', 'AXTitle', title, recursive=True)

    def sheets(self, role_description: Optional[str] = None):
        return self._convenience_match('AXSheet', 'AXDescription', role_description)

    def sheets_r(self, role_description: Optional[str] = None):
        return self._convenience_match('AXSheet', 'AXDescription', role_description, recursive=True)

    def static_texts(self, value: Optional[str] = None):
        return self._convenience_match('AXStaticText', 'AXValue', value)

    def static_texts_r(self, value: Optional[str] = None):
        return self._convenience_match('AXStaticText', 'AXValue', value, recursive=True)

    def groups(self, role_description: Optional[str] = None):
        """Return a list of groups with an optional match parameter."""
        return self._convenience_match('AXGroup', 'AXRoleDescription', role_description)

    def groups_r(self, role_description: Optional[str] = None):
        """Return a list of groups with an optional match parameter."""
        return self._convenience_match('AXGroup', 'AXRoleDescription', role_description, recursive=True)

    def radio_buttons(self, title: Optional[str] = None):
        """Return a list of radio buttons with an optional match parameter."""
        return self._convenience_match('AXRadioButton', 'AXTitle', title)

    def radio_buttons_r(self, title: Optional[str] = None):
        """Return a list of radio buttons with an optional match parameter."""
        return self._convenience_match('AXRadioButton', 'AXTitle', title, recursive=True)

    def pop_up_buttons(self, title: Optional[str] = None):
        """Return a list of popup menus with an optional match parameter."""
        return self._convenience_match('AXPopUpButton', 'AXTitle', title)

    def pop_up_buttons_r(self, title: Optional[str] = None):
        """Return a list of popup menus with an optional match parameter."""
        return self._convenience_match('AXPopUpButton', 'AXTitle', title, recursive=True)

    def rows(self, title: Optional[str] = None):
        """Return a list of rows with an optional match parameter."""
        return self._convenience_match('AXRow', 'AXTitle', title)

    def rows_r(self, title: Optional[str] = None):
        """Return a list of rows with an optional match parameter."""
        return self._convenience_match('AXRow', 'AXTitle', title, recursive=True)

    def sliders(self, value: Optional[str] = None):
        """Return a list of sliders with an optional match parameter."""
        return self._convenience_match('AXSlider', 'AXValue', value)

    def sliders_r(self, value: Optional[str] = None):
        """Return a list of sliders with an optional match parameter."""
        return self._convenience_match('AXSlider', 'AXValue', value, recursive=True)

    @staticmethod
    def menu_items(menuitem, *args):
        """ Return the specified menu item.
            Refer to items by name:
                app._menuItem(app.AXMenuBar, 'File', 'New').Press()
                app._menuItem(app.AXMenuBar, 'Edit', 'Insert', 'Line Break').Press()
            Refer to items by index:
                app._menuitem(app.AXMenuBar, 1, 0).Press()
            Refer to items by mix-n-match:
                app._menuitem(app.AXMenuBar, 1, 'About TextEdit').Press()
        """
        for item in args:
            if menuitem.AXChildren[0].AXRole == 'AXMenu':  # If the item has an AXMenu as a child, navigate into it.
                menuitem = menuitem.AXChildren[0]
            try:
                menuitem = menuitem.AXChildren[int(item)]  # Find AXMenuBarItems and AXMenuItems using a handy wildcard.
            except ValueError:
                menuitem = menuitem.find_first(AXRole='AXMenu*Item', AXTitle=item)
        return menuitem

    def _convenience_match(self, role, attr, match, recursive=False):
        """Method used by role based convenience functions to find a match."""
        kwargs = dict()
        if match:  # If the user supplied some text to search for, supply that in the kwargs.
            kwargs[attr] = match
        return self._find_elements(AXRole=role, recursive=recursive, **kwargs)

    def _find_element(self, recursive: bool = False, **kwargs):
        """Return the first object that matches the criteria."""
        for item in self._find_elements(recursive=recursive, **kwargs):
            return item

    def _find_elements(self, recursive: bool = False, **kwargs):
        """Return a list of all children that match the specified criteria."""
        return list(filter(match_filter(**kwargs), self._generate_children(recursive=recursive), ))

    def _generate_children(self, target=None, recursive=False):
        """Generator which yields all AXChildren of the object."""
        if target is None:
            target = self

        if 'AXChildren' not in target.ax_attributes:
            return

        for child in target.AXChildren:
            yield child
            if recursive:
                for c in self._generate_children(child, recursive):
                    yield c


def match_filter(**kwargs):
    def _match(obj):
        for k in kwargs.keys():
            try:
                val = getattr(obj, k)
            except AttributeError:
                return False
            if isinstance(val, str):
                if not fnmatch.fnmatch(val, kwargs[k]):
                    return False
            else:
                if val != kwargs[k]:
                    return False
        return True

    return _match
